#!/usr/bin/env python3
import json
from optparse import OptionParser
import os.path
import polib
import shutil
import sys


parser = OptionParser()
parser.add_option("-j", "--json-output", dest="output_file",
                  help="Write result in structured JSON format to file")
(options, args) = parser.parse_args()


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def read_all_messages(path):
    if not os.path.isfile(path):
        raise Exception("cannot read {}".format(path))
    pofile = polib.pofile(path)
    messages = set()
    for entry in pofile.untranslated_entries():
        messages.add(entry.msgid)
        if entry.msgid_plural:
            messages.add(entry.msgid_plural)
    return messages


def compare_po(old_pot_path, new_pot_path):
    print(f"Reading {old_pot_path}")
    old_messages = read_all_messages(old_pot_path)
    print(f"{len(old_messages)} message(s) read from {old_pot_path}")
    print()

    print(f"Reading {new_pot_path}")
    new_messages = read_all_messages(new_pot_path)
    print(f"{len(new_messages)} message(s) read from {new_pot_path}")
    print()

    print("Computing differences...")
    deleted_messages = list(sorted(old_messages - new_messages))
    added_messages = list(sorted(new_messages - old_messages))
    print()
    return (deleted_messages, added_messages)


def output_to_screen(deleted_messages, added_messages):
    columns = shutil.get_terminal_size((80, 24)).columns
    delim = "-" * columns

    if len(deleted_messages) == 0:
        print("No message deleted.")
    else:
        print(f"{len(deleted_messages)} message(s) deleted:")
        for msg in deleted_messages:
            print(f"{bcolors.OKCYAN}{delim}{bcolors.ENDC}")
            print(f"{bcolors.OKCYAN}{msg}{bcolors.ENDC}")
        print(f"{bcolors.OKCYAN}{delim}{bcolors.ENDC}")
    print()

    if len(added_messages) == 0:
        print("No message added.")
    else:
        print(f"{len(added_messages)} message(s) added:")
        for msg in added_messages:
            print(f"{bcolors.OKGREEN}{delim}{bcolors.ENDC}")
            print(f"{bcolors.OKGREEN}{msg}{bcolors.ENDC}")
        print(f"{bcolors.OKGREEN}{delim}{bcolors.ENDC}")


def output_to_json(path, deleted_messages, added_messages):
    result = {"deleted": deleted_messages, "added": added_messages}
    with open(path, "w") as fp:
        fp.write(json.dumps(result))


def main():
    if len(sys.argv) < 3:
        print("Usage: python3 ./tools/pot_diff.py <old.pot> <new.pot>")
        return 1
    (deleted_messages, added_messages) = compare_po(sys.argv[1], sys.argv[2])
    if options.output_file:
        output_to_json(options.output_file, deleted_messages, added_messages)
        print("Comparison result written to {}".format(options.output_file))
    else:
        output_to_screen(deleted_messages, added_messages)
    return 0


exit(main())



#!/usr/bin/env python3
"""Search for matches within the json data.

Example usages:

    %(prog)s --all type=dream strength=2

    %(prog)s material=plastic material=steel

"""

import sys
import argparse
from util import import_data, matches_all_wheres, CDDAJSONWriter, WhereAction

parser = argparse.ArgumentParser(
    description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
parser.add_argument(
    "--fnmatch",
    default="*.json",
    help="override with glob expression to select a smaller fileset.")
parser.add_argument(
    "--all",
    action="store_true",
    help="includes all matches. by default, include only the first match.")
parser.add_argument(
    "where",
    action=WhereAction, nargs='+', type=str,
    help="where exclusions of the form 'where_key=where_val', no quotes.")


if __name__ == "__main__":
    args = parser.parse_args()

    json_data, load_errors = import_data(json_fmatch=args.fnmatch)
    if load_errors:
        # If we start getting unexpected JSON or other things, might need to
        # revisit quitting on load_errors
        print("Error loading JSON data.")
        for e in load_errors:
            print(e)
        sys.exit(1)
    elif not json_data:
        print("No data loaded.")
        sys.exit(1)

    # Wasteful iteration, but less code to maintain on a tool that will likely
    # change again.
    plucked = [
        item for item in json_data if matches_all_wheres(item, args.where)]

    if not plucked:
        print("Nothing found.")
        sys.exit(1)
    elif plucked and not args.all:
        print(CDDAJSONWriter(plucked[0]).dumps())
    else:
        # ugh, didn't realize JSON writer only wanted single objects when I
        # wrote it.
        # TODO: get rid of ugh
        print("[")
        for i, p in enumerate(plucked):
            eol = ",\n" if i < len(plucked) - 1 else "\n"
            print(CDDAJSONWriter(p, 1).dumps(), end=eol)
        print("]")



#!/usr/bin/env python3

import argparse
import json
import os

args = argparse.ArgumentParser()
args.add_argument("dir", action="store", help="specify json directory")
args_dict = vars(args.parse_args())


def gen_new(path):
    change = False
    with open(path, "r", encoding="utf-8") as json_file:
        json_data = json.load(json_file)
        for jo in json_data:
            if (type(jo) is dict and "type" in jo and
                    jo["type"] == "MONSTER_FACTION"):
                if "by_mood" in jo and type(jo["by_mood"]) is str:
                    jo["by_mood"] = [jo.pop("by_mood")]
                    change = True
                if "neutral" in jo and type(jo["neutral"]) is str:
                    jo["neutral"] = [jo.pop("neutral")]
                    change = True
                if "friendly" in jo and type(jo["friendly"]) is str:
                    jo["friendly"] = [jo.pop("friendly")]
                    change = True
                if "hate" in jo and type(jo["hate"]) is str:
                    jo["hate"] = [jo.pop("hate")]
                    change = True

    return json_data if change else None


for root, directories, filenames in os.walk(args_dict["dir"]):
    for filename in filenames:
        path = os.path.join(root, filename)
        if path.endswith(".json"):
            new = gen_new(path)
            if new is not None:
                with open(path, "w", encoding="utf-8") as jf:
                    json.dump(new, jf, ensure_ascii=False)
                os.system(f"./tools/format/json_formatter.cgi {path}")



from ..write_text import write_text


def parse_weakpoint_set(json, origin):
    if "weakpoints" in json:
        for wp in json["weakpoints"]:
            id = ""
            if "id" in wp:
                id = wp["id"]
            if "name" in wp:
                comment = "Sentence fragment describing " \
                          "the \"{}\" weakpoint".format(id)
                write_text(wp["name"], origin, comment=comment)
            if "effects" in wp:
                for fx in wp["effects"]:
                    if "message" in fx:
                        comment = "Message describing the effect of hitting " \
                                  "the \"{}\" weakpoint".format(id)
                        write_text(fx["message"], origin, comment=comment)



#!/usr/bin/env python3
from . import DefaultKnownWords, Tokenizer
import os
import polib
import unicodedata


def gen_dictionary():
    po_path = os.path.join(os.path.dirname(__file__),
                           "../../lang/po/cataclysm-dda.pot")
    pofile = polib.pofile(po_path)
    dictionary = set()
    for entry in pofile:
        words = Tokenizer.findall(entry.msgid)
        for word in words:
            norm = unicodedata.normalize('NFC', word)
            if norm not in DefaultKnownWords:
                dictionary.add(norm)
        if entry.msgid_plural:
            words = Tokenizer.findall(entry.msgid_plural)
            for word in words:
                norm = unicodedata.normalize('NFC', word)
                if norm not in DefaultKnownWords:
                    dictionary.add(norm)
    dict_path = os.path.join(os.path.dirname(__file__), "dictionary.txt")
    with open(dict_path, "w", encoding="utf-8") as fp:
        for word in sorted(dictionary, key=lambda v: (v.lower(), v)):
            print(word, file=fp)



#!/usr/bin/env python3
import polib
import os
import re
import sys
types = ["d", "i", "u", "f", "e", "g", "x", "s", "c", "p"]


def findall(p, s):
    i = s.find(p)
    while i != -1:
        yield i
        i = s.find(p, i + 1)


def is_formatted_entry(msg):
    if "%" not in msg:
        return False
    indices = list(findall("%", msg))
    index_iter = iter(indices)
    for i in index_iter:
        if i + 1 == len(msg):
            return False
        if msg[i + 1] == "%":  # skip %%
            next(index_iter, None)
            continue
        if not msg[i + 1].isnumeric() and \
           msg[i + 1] not in ["d", "s", "c", "l", "f", "."]:
            return False
        if msg[i + 1] == "." and \
           (i + 2 == len(msg) or not msg[i + 2].isnumeric()):
            return False
    return True


def decompose_format_strings(msg):
    indices = list(findall("%", msg))
    non_positional = []
    positional = []
    segments = []
    index_iter = iter(indices)
    for i in index_iter:
        if i + 1 == len(msg):
            break
        if msg[i + 1] == "%":  # skip %%
            next(index_iter, None)
            continue
        if msg[i + 1].isnumeric() and msg[i + 2] == "$":
            # positional format strings %1$s, %2$d, etc.
            positional.append(msg[i:i + 4])
            segments.append((i, i + 4))
        else:
            # non-positional format strings %s, %3d, etc.
            idx = i + 1
            while msg[idx] not in types and idx + 1 < len(msg):
                idx += 1
            if idx == len(msg):
                break
            # ignore the cases where "%d" is translated to "%2d"
            non_positional.append(re.sub(r'[0-9]+', '', msg[i:idx + 1]))
            segments.append((i, idx + 1))
    return (non_positional, positional, segments)


def get_type(arg):
    answer = ""
    for i in range(len(arg) - 1, -1, -1):
        if arg[i].isnumeric() or arg[i] == '.' \
           or arg[i] == '$' or arg[i] == '%':
            break
        answer = arg[i] + answer
    return answer


def check_message(entry):
    msgid = entry.msgid
    msgstr = entry.msgstr
    if not is_formatted_entry(msgid) and not is_formatted_entry(msgstr):
        return (True, [], [], "")
    (non_pos_msgid, pos_msgid, seg_msgid) = decompose_format_strings(msgid)
    if len(non_pos_msgid) > 0 and len(pos_msgid) > 0:
        return (False, seg_msgid, [],
                "Cannot mix positional and non-positional arguments"
                " in format string")
    (non_pos_msgstr, pos_msgstr, seg_msgstr) = decompose_format_strings(msgstr)
    if len(non_pos_msgstr) > 0 and len(pos_msgstr) > 0:
        return (False, [], seg_msgstr,
                "Cannot mix positional and non-positional arguments"
                " in format string")
    if non_pos_msgid == non_pos_msgstr and \
       sorted(pos_msgid) == sorted(pos_msgstr):
        return (True, [], [], "")
    # "%2$d %1$s" is considered equivalent to "%s %d"
    # as the order of types is preserved
    msgid_types = []
    msgstr_types = []
    for arg in non_pos_msgid + sorted(pos_msgid):
        msgid_types.append(get_type(arg))
    for arg in non_pos_msgstr + sorted(pos_msgstr):
        msgstr_types.append(get_type(arg))
    if len(msgid_types) != len(msgstr_types):
        return (False, seg_msgid, seg_msgstr, "The number of arguments differ")
    for i in range(0, len(msgid_types)):
        if msgid_types[i] != msgstr_types[i]:
            return (False, seg_msgid, seg_msgstr,
                    "The types of arguments differ")
    return (True, [], [], "")


def check_po_file(file):
    pofile = polib.pofile(file)
    errors = []
    for entry in pofile.translated_entries():
        if entry.msgid_plural:
            # TODO: implement proper check for plural messages
            continue
        (ok, seg_msgid, seg_msgstr, reason) = check_message(entry)
        if not ok:
            errors.append((entry.msgid, entry.msgstr,
                           seg_msgid, seg_msgstr, reason))
    return errors


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_message(msg, segments):
    beg = [x for (x, y) in segments]
    end = [y for (x, y) in segments]
    for i in range(len(msg)):
        if i in beg:
            print(bcolors.FAIL, end="")
            print(bcolors.UNDERLINE, end="")
        elif i in end:
            print(bcolors.ENDC, end="")
        print(msg[i], end="")
    print(bcolors.ENDC)


po_files = []
for file in sorted(os.listdir("lang/po")):
    if file.endswith(".po") and not file.endswith("en.po"):
        po_files.append(file)
files_to_check = []
if len(sys.argv) == 1:
    files_to_check = po_files
else:
    for i in range(1, len(sys.argv)):
        if sys.argv[i] + ".po" in po_files:
            files_to_check.append(sys.argv[i] + ".po")
        else:
            print("Warning: Unknown language", sys.argv[i])
num_errors = 0
for file in sorted(files_to_check):
    print("Checking", file, end="", flush=True)
    errors = check_po_file("lang/po/" + file)
    n = len(errors)
    num_errors += n
    if n > 0:
        print(f" => {bcolors.FAIL}{n} error(s) detected:{bcolors.ENDC}")
        for (msgid, msgstr, seg_msgid, seg_msgstr, reason) in errors:
            print(f"{bcolors.BOLD}problem   :{bcolors.ENDC}", end="")
            print(reason)
            print(f"{bcolors.BOLD}original  :{bcolors.ENDC}", end="")
            print_message(msgid.replace("\n\n", "\n"), seg_msgid)
            print(f"{bcolors.BOLD}translated:{bcolors.ENDC}", end="")
            print_message(msgstr.replace("\n\n", "\n"), seg_msgstr)
            print()
    else:
        print(f" => {bcolors.OKGREEN}No error detected.{bcolors.ENDC}")
    print()
exit(num_errors)



from dataclasses import dataclass


@dataclass
class Message:
    comments: list
    origin: str
    format_tag: str
    context: str
    text: str
    text_plural: str
    explicit_plural: bool


messages = dict()
occurrences = []



from .effect import parse_effect_on_condition
from ..write_text import write_text


def extract_eoc_id(json):
    if type(json) is str:
        return json  # eoc id
    elif type(json) is dict and "id" in json:
        return json.get("id")  # eoc defined inline
    else:
        raise Exception("result_eocs member missing id")


def result_hint(json):
    hint = json.get("result", json.get("result_eocs", None))
    if hint is None:
        raise Exception("no result hint for recipe translation,"
                        " needs 'result' or 'result_eocs' defined")
    if type(hint) is str:
        return hint
    elif type(hint) is list:
        return ", ".join(map(extract_eoc_id, hint))
    else:
        raise Exception("recipe's result_eocs has unsupported form")


def parse_recipe(json, origin):
    if "book_learn" in json and type(json["book_learn"]) is dict:
        for book in json["book_learn"]:
            if "recipe_name" in json["book_learn"][book]:
                write_text(json["book_learn"][book]["recipe_name"], origin,
                           comment="Recipe name learnt from book")
    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of recipe crafting \"{}\""
                   .format(result_hint(json)))
    if "name" in json:
        write_text(json["name"], origin,
                   comment="Custom name for recipe crafting \"{}\""
                   .format(result_hint(json)))
    if "blueprint_name" in json:
        write_text(json["blueprint_name"], origin,
                   comment="Blueprint name of recipe crafting \"{}\""
                   .format(result_hint(json)))
    if "result_eocs" in json:
        for eoc in json["result_eocs"]:
            if type(eoc) is dict:
                parse_effect_on_condition(eoc, origin,
                                          "result of recipe crafting \"{}\""
                                          .format(result_hint(json)))



from ..write_text import write_text
from .effect import parse_effect_on_condition


def parse_examine_action(json, origin, name):
    ename = "{}'s examine action".format(name)
    if "query_msg" in json:
        write_text(json["query_msg"], origin,
                   comment="Query message for {}".format(ename))
    if "success_msg" in json:
        write_text(json["success_msg"], origin,
                   comment="Message when {} succeeds".format(ename))
    if "redundant_msg" in json:
        write_text(json["redundant_msg"], origin,
                   comment="Message when {} would be redundant".format(ename))
    if "condition_fail_msg" in json:
        write_text(json["condition_fail_msg"], origin,
                   comment="Message when {} fails its condition".format(ename))
    if "effect_on_conditions" in json:
        for eoc in json["effect_on_conditions"]:
            if type(eoc) is dict:
                parse_effect_on_condition(eoc, origin, ename)



from ..helper import get_singular_name
from ..write_text import write_text


def parse_vehicle_part_category(json, origin):
    name = get_singular_name(json["name"])
    write_text(json["name"], origin, comment="Name of a vehicle part category")
    write_text(json["short_name"], origin,
               comment="Short name of vehicle part category \"{}\""
               .format(name))



from ..helper import get_singular_name
from ..write_text import write_text
from .examine_action import parse_examine_action


def parse_terrain(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Terrain name")
    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of terrain \"{}\"".format(name))
    if "bash" in json:
        if "sound" in json["bash"]:
            write_text(json["bash"]["sound"], origin,
                       comment="Bashing sound")
        if "sound_fail" in json["bash"]:
            write_text(json["bash"]["sound_fail"], origin,
                       comment="Bashing failed sound")
    if "boltcut" in json:
        if "message" in json["boltcut"]:
            write_text(json["boltcut"]["message"], origin,
                       comment="Boltcut message of terrain \"{}\"".
                       format(name))
        if "sound" in json["boltcut"]:
            write_text(json["boltcut"]["sound"], origin,
                       comment="Boltcut sound of terrain \"{}\"".format(name))
    if "hacksaw" in json:
        if "message" in json["hacksaw"]:
            write_text(json["hacksaw"]["message"], origin,
                       comment="Hacksaw message of terrain \"{}\""
                       .format(name))
        if "sound" in json["hacksaw"]:
            write_text(json["hacksaw"]["sound"], origin,
                       comment="Hacksaw sound of terrain \"{}\"".format(name))
    if "prying" in json:
        if "message" in json["prying"]:
            write_text(json["prying"]["message"], origin,
                       comment="Prying message of terrain \"{}\"".format(name))
        if "prying_data" in json["prying"]:
            if "failure" in json["prying"]["prying_data"]:
                write_text(json["prying"]["prying_data"]["failure"], origin,
                           comment="Prying failure of terrain \"{}\""
                           .format(name))
    if "lockpick_message" in json:
        write_text(json["lockpick_message"], origin,
                   comment="Lockpick message of terrain \"{}\"".format(name))
    if "examine_action" in json:
        parse_examine_action(json["examine_action"], origin,
                             "terrain \"{}\"".format(name))



from ..write_text import write_text


def parse_damage_info_order(json, origin):
    if "verb" in json:
        write_text(json["verb"], origin,
                   comment="A verb describing how this damage type"
                           "is applied (ex: \"bashing\")")



from ..write_text import write_text


def parse_gate(json, origin):
    messages = json.get("messages", {})

    for i in messages:
        write_text(messages[i], origin,
                   comment="Message of {} action on a gate".format(i))



import codecs
from datetime import datetime, timezone
import json
import os
import polib

from .message import messages, occurrences


def process_comments(comments, origins, obsolete_paths):
    result = []

    # remove duplicate comment lines while preserving order
    seen = set()
    for comment in comments:
        if comment not in seen:
            result.append(comment)
            seen.add(comment)

    # add 'obsolete' comment if this string
    # has only obsolete files in the origins
    obsolete_count = 0
    for origin in origins:
        is_obsolete = False

        if "obsolet" in origin:
            # if the file path contains "obsolete"/"obsoletion"
            is_obsolete = True
        else:
            # if the file path matches the obsolete paths
            # explicitly specified in the '-D' arguments
            for o_path in obsolete_paths:
                p = os.path.commonpath([o_path, origin])
                if p in obsolete_paths:
                    is_obsolete = True
                    break
        if is_obsolete:
            obsolete_count += 1

    if obsolete_count == len(origins):
        return [
            "[DEPRECATED] Don't translate this line.",
            "This string is from an obsolete source "
            "and will no longer be used in the game."
        ]
    return result


def sanitize(reference, pkg_name="Cataclysm-DDA"):
    if not os.path.isfile(reference):
        raise Exception(f"Cannot read {reference}")
    pofile = polib.pofile(reference)

    # sanitize plural entries
    # Multiple objects may define slightly different plurals for strings,
    # but without the specified context only one such string can be stored.
    # Adds a plural form to those matching strings that do not have it.
    for entry in pofile.untranslated_entries():
        pair = (entry.msgctxt if entry.msgctxt else "", entry.msgid)
        if pair in messages:
            # first, check the messages against the reference
            for m in messages[pair]:
                if not m.text_plural and entry.msgid_plural:
                    m.text_plural = entry.msgid_plural
                    break
            # then check the reference against the messages
            # prioritize plurals that are explicitly specified in JSON
            temp_plural = ""
            for m in messages[pair]:
                if m.text_plural and not entry.msgid_plural:
                    entry.msgstr_plural = {0: "", 1: ""}
                    if m.explicit_plural:
                        entry.msgid_plural = m.text_plural
                        break
                    temp_plural = m.text_plural
            if temp_plural:
                entry.msgid_plural = temp_plural

    # write the correct header.
    tzinfo = datetime.now(timezone.utc).astimezone().tzinfo
    tztime = datetime.now(tzinfo).strftime('%Y-%m-%d %H:%M%z')
    pofile.metadata = {
        "Project-Id-Version": pkg_name,
        "POT-Creation-Date": f"{tztime}",
        "PO-Revision-Date": f"{tztime}",
        "Last-Translator": "None",
        "Language-Team": "None",
        "Language": "en",
        "MIME-Version": "1.0",
        "Content-Type": "text/plain; charset=UTF-8",
        "Content-Transfer-Encoding": "8bit",
        "Plural-Forms": "nplurals=2; plural=(n > 1);"
    }
    pofile.metadata_is_fuzzy = 0

    pofile.save()


def is_unicode(sequence):
    hex = "0123456789abcdef"
    return sequence[0] == "\\" and sequence[1] == "u" and \
        sequence[2] in hex and sequence[3] in hex and \
        sequence[4] in hex and sequence[5] in hex


def restore_unicode(string):
    i = string.find("\\u")
    while i != -1 and i + 5 < len(string):
        slice = string[i:i + 6]
        if is_unicode(slice):
            string = string.replace(slice,
                                    codecs.unicode_escape_decode(slice)[0])
        i = string.find("\\u")
    return string


def format_msg(text):
    return restore_unicode(json.dumps(text))


def write_to_pot(fp, obsolete_paths=[]):
    entries = []
    for (context, text) in occurrences:
        if (context, text) not in messages:
            continue
        comments = []
        origins = set()
        format_tag = ""
        text_plural = ""
        for message in messages[(context, text)]:
            comments = comments + message.comments
            origins.add(message.origin)
            if message.format_tag:
                format_tag = message.format_tag
            if message.text_plural:
                text_plural = message.text_plural
        origin = " ".join(sorted(origins))
        entry = []

        # translator comments
        for line in process_comments(comments, origins, obsolete_paths):
            entry.append(f"#. ~ {line}")

        # reference
        entry.append(f"#: {origin}")

        # c-format
        if format_tag:
            entry.append(f"#, {format_tag}")

        # context
        if context:
            entry.append(f"msgctxt \"{context}\"")

        # text
        if text_plural:
            entry.append(f"msgid {format_msg(text)}\n"
                         f"msgid_plural {format_msg(text_plural)}\n"
                         "msgstr[0] \"\"\n"
                         "msgstr[1] \"\"")
        else:
            entry.append(f"msgid {format_msg(text)}\n"
                         "msgstr \"\"")

        entries.append("\n".join(entry))
        del messages[(context, text)]

    fp.write("\n\n".join(entries))



from ..write_text import write_text


def parse_item_action(json, origin):
    if "name" in json:
        write_text(json["name"], origin,
                   comment="Item action name of \"{}\"".format(json["id"]))



#!/usr/bin/env python3

import argparse
import json
import os

args = argparse.ArgumentParser()
args.add_argument("dir", action="store", help="specify json directory")
args_dict = vars(args.parse_args())


def to_damage_vals(jo):
    convert_stab = False
    if "flags" in jo and ("STAB" in jo["flags"] or "SPEAR" in jo["flags"]):
        convert_stab = True
        if "STAB" in jo["flags"]:
            jo["flags"].remove("STAB")

    res = dict()
    if "bashing" in jo:
        res["bash"] = jo.pop("bashing")
    if "cutting" in jo:
        cut_str = "cut"
        if convert_stab:
            cut_str = "stab"
        res[cut_str] = jo.pop("cutting")

    if res:
        jo["melee_damage"] = res

    if "proportional" in jo:
        res = dict()
        if "bashing" in jo["proportional"]:
            res["bash"] = jo["proportional"].pop("bashing")
        if "cutting" in jo["proportional"]:
            cut_str = "cut"
            if convert_stab:
                cut_str = "stab"
            res[cut_str] = jo["proportional"].pop("cutting")
        if res:
            jo["proportional"]["melee_damage"] = res

    if "relative" in jo:
        res = dict()
        if "bashing" in jo["relative"]:
            res["bash"] = jo["relative"].pop("bashing")
        if "cutting" in jo["relative"]:
            cut_str = "cut"
            if convert_stab:
                cut_str = "stab"
            res[cut_str] = jo["relative"].pop("cutting")
        if res:
            jo["relative"]["melee_damage"] = res

    return jo


def gen_new(path):
    change = False
    # Having troubles with this script halting?
    # Uncomment below to find the file it's halting on
    #print(path)
    with open(path, "r", encoding="utf-8") as json_file:
        try:
            json_data = json.load(json_file)
        except json.JSONDecodeError:
            print(f"Json Decode Error at {path}")
            print("Ensure that the file is a JSON file consisting of"
                  " an array of objects!")
            return None

        for jo in json_data:
            if type(jo) is not dict:
                print(f"Incorrectly formatted JSON data at {path}")
                print("Ensure that the file is a JSON file consisting"
                      " of an array of objects!")
                break

            item_types = {
                "GENERIC",
                "AMMO",
                "GUN",
                "ARMOR",
                "PET_ARMOR",
                "TOOL",
                "TOOLMOD",
                "TOOL_ARMOR",
                "BOOK",
                "COMESTIBLE",
                "ENGINE",
                "WHEEL",
                "GUNMOD",
                "MAGAZINE",
                "BATTERY",
                "BIONIC_ITEM"
            }

            if ("type" not in jo or jo["type"] not in item_types):
                continue

            new_data = to_damage_vals(jo)
            if new_data is not None:
                jo = new_data
                change = True

    return json_data if change else None


for root, directories, filenames in os.walk(args_dict["dir"]):
    for filename in filenames:
        if not filename.endswith(".json"):
            continue

        path = os.path.join(root, filename)
        new = gen_new(path)
        if new is not None:
            with open(path, "w", encoding="utf-8") as jf:
                json.dump(new, jf, ensure_ascii=False)
            os.system(f"./tools/format/json_formatter.cgi {path}")



from ..helper import get_singular_name
from ..write_text import write_text


def parse_achievement(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="Name of achievement")
    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of achievement \"{}\"".format(name))
    if "requirements" in json:
        for req in json["requirements"]:
            if "description" in req:
                comment = "Description of requirement " \
                          "of achievement \"{}\"".format(name)
                write_text(req["description"], origin, comment=comment)



from ..helper import get_singular_name
from ..write_text import write_text


def parse_fault_fix(json, origin):
    name = get_singular_name(json["name"])
    write_text(json["name"], origin,
               comment="Name of fault_fix \"{}\"".format(name))
    if "success_msg" not in json:
        return
    write_text(json["success_msg"], origin,
               comment="Success message of fault_fix \"{}\"".format(name))



from ..write_text import write_text


# See `npc_template::load` in `npc.cpp`.
chatbin_snippets = [
    "<acknowledged>"
    "<camp_food_thanks>"
    "<camp_larder_empty>"
    "<camp_water_thanks>"
    "<cant_flee>"
    "<close_distance>"
    "<combat_noise_warning>"
    "<danger_close_distance>"
    "<done_mugging>"
    "<far_distance>"
    "<fire_bad>"
    "<fire_in_the_hole_h>"
    "<fire_in_the_hole>"
    "<general_danger_h>"
    "<general_danger>"
    "<heal_self>"
    "<hungry>"
    "<im_leaving_you>"
    "<its_safe_h>"
    "<its_safe>"
    "<keep_up>"
    "<kill_npc_h>"
    "<kill_npc>"
    "<kill_player_h>"
    "<let_me_pass>"
    "<lets_talk>"
    "<medium_distance>"
    "<monster_warning_h>"
    "<monster_warning>"
    "<movement_noise_warning>"
    "<need_batteries>"
    "<need_booze>"
    "<need_fuel>"
    "<no_to_thorazine>"
    "<run_away>"
    "<speech_warning>"
    "<thirsty>"
    "<wait>"
    "<warn_sleep>"
    "<yawn>"
    "<yes_to_lsd>"
    "snip_pulp_zombie"
    "snip_heal_player"
    "snip_mug_dontmove"
    "snip_wound_infected"
    "snip_wound_bite"
    "snip_radiation_sickness"
    "snip_bleeding"
    "snip_bleeding_badly"
    "snip_lost_blood"
    "snip_bye"
    "snip_consume_cant_accept"
    "snip_consume_cant_consume"
    "snip_consume_rotten"
    "snip_consume_eat"
    "snip_consume_need_item"
    "snip_consume_med"
    "snip_consume_nocharge"
    "snip_consume_use_med"
    "snip_give_nope"
    "snip_give_to_hallucination"
    "snip_give_cancel"
    "snip_give_dangerous"
    "snip_give_wield"
    "snip_give_weapon_weak"
    "snip_give_carry"
    "snip_give_carry_cant"
    "snip_give_carry_cant_few_space"
    "snip_give_carry_cant_no_space"
    "snip_give_carry_too_heavy"
    "snip_wear"
]


def parse_npc(json, origin):
    gender = "an"
    if "gender" in json:
        gender = "a {}".format(json["gender"])
    comment = json.get("//", None)
    if "name_unique" in json:
        write_text(json["name_unique"], origin,
                   comment=["Unique name of {} NPC".format(gender), comment])
    if "name_suffix" in json:
        write_text(json["name_suffix"], origin,
                   comment=["Name suffix of {} NPC".format(gender), comment])
    for snip in chatbin_snippets:
        if snip in json:
            write_text(json[snip], origin,
                       comment=["Chatbin snippet {} of {} NPC"
                                .format(snip, gender), comment])



#!/usr/bin/env python3
import polib
import os
import re
import sys


tags = set()
ignore_tags = {'<1>', '<2>', '<3>', '<4>', '<f>', '<n>',
               '<q>', '<r>', '<s>', '<u>', '<more>', '<empty>'}
pattern = re.compile(r'<[a-z/0-9_]+>')


def init_tags():
    global tags
    pofile = polib.pofile("./lang/po/cataclysm-dda.pot")
    for entry in pofile:
        for tag in set(pattern.findall(entry.msgid)):
            tags.add(tag)
    tags = tags - ignore_tags


def extract_tags(msg):
    matches = pattern.findall(msg)
    result = set()
    for match in matches:
        if match in tags:
            result.add(match)
    return result


def check_message(entry):
    msgid = entry.msgid
    msgstr = entry.msgstr
    if not msgstr:
        return set()
    tags_msgid = extract_tags(msgid)
    tags_msgstr = extract_tags(msgstr)
    return tags_msgid - tags_msgstr


def check_po_file(file):
    pofile = polib.pofile(file)
    errors = 0
    for entry in pofile.translated_entries():
        missing_tags = check_message(entry)
        if missing_tags:
            print("Tag(s) {} missing in translation: {} => {}".format(
                missing_tags,
                entry.msgid.replace("\n", "\\n"),
                entry.msgstr.replace("\n", "\\n")))
            errors += 1
    return errors


init_tags()
po_files = []
for file in sorted(os.listdir("lang/po")):
    if file.endswith(".po") and not file.endswith("en.po"):
        po_files.append(file)
files_to_check = []
if len(sys.argv) == 1:
    files_to_check = po_files
else:
    for i in range(1, len(sys.argv)):
        if sys.argv[i] + ".po" in po_files:
            files_to_check.append(sys.argv[i] + ".po")
        else:
            print("Warning: Unknown language", sys.argv[i])
num_errors = 0
for file in sorted(files_to_check):
    print("Checking {}".format(file))
    num_errors += check_po_file("lang/po/" + file)
    print()
exit(num_errors)



#!/usr/bin/env python3
"""
Outer join game IDs with overlay IDs and tileset IDs


cd Cataclysm-DDA

python3 tools/json_tools/table.py -f csv --nonestring ""\
 --tileset type id name description color looks_like "copy-from" longest_side \
 > all_game_ids.csv

python3 tools/gfx_tools/list_tileset_ids.py gfx/UltimateCataclysm \
 > tileset_ids.csv

python3 tools/json_tools/generate_overlay_ids.py \
 > all_overlay_ids.csv

python3 tools/gfx_tools/tileset_coverage.py \
 all_game_ids.csv all_overlay_ids.csv tileset_ids.csv \
 tileset_coverage_output.csv
"""
import argparse
import re
import sys

from typing import Union

import pandas


REPLACEMENTS = (
    r'^overlay_(.+)_sees_player$',
)

DELETION_RE = (
    r'_season_(spring|summer|autumn|winter)|'
    r'^\[|'
    r'\]$|'
    r'^overlay(_(male|female))?_(effect|mutation|worn|wielded(_corpse)?)_|'
    r'^overlay_|'
    r'^corpse_|'
    r'^vp_|'
    r'(_('
    r'cover|cross|horizontal|horizontal_2|vertical|vertical_2|ne|nw|se|sw'
    r'))?'
    r'(_('
    r'unconnected|left|right|rear|front'
    r'))?'
    r'(_edge)?'
    r'$'
)


def strip_overlay_id(overlay_id: str) -> str:
    """
    Extract game ID from an overlay ID string

    >>> strip_overlay_id('vp_wheel_wood')
    'wheel_wood'
    >>> strip_overlay_id('vp_reinforced_windshield_front_edge')
    'reinforced_windshield'
    >>> strip_overlay_id('vp_wheel_wheelchair')
    'wheel_wheelchair'
    >>> strip_overlay_id('vp_frame_wood_vertical_2_unconnected')
    'frame_wood'

    """
    stripped_id = overlay_id

    for pattern in REPLACEMENTS:
        stripped_id = re.sub(
            pattern,
            r'\g<1>',
            overlay_id,
        )

    stripped_id = re.sub(
        DELETION_RE,
        '',
        stripped_id,
    )

    return stripped_id


def get_data(
        all_game_ids_filename: str,
        overlay_ids_filename: str,
        tileset_ids_filename: str)\
        -> tuple:
    """
    Load datasets with game IDs, tileset IDs and overlay IDs
    """
    all_game_ids = pandas.read_csv(
        all_game_ids_filename,
        warn_bad_lines=True,
    )
    overlay_ids = pandas.read_csv(
        overlay_ids_filename,
        header=None,
        names=('overlay_id',),
        warn_bad_lines=True,
    )
    tileset_ids = pandas.read_csv(
        tileset_ids_filename,
        header=None,
        names=('tileset_id',),
        warn_bad_lines=True,
    )
    return all_game_ids, overlay_ids, tileset_ids


def merge_datasets(
        all_game_ids: Union[dict, pandas.DataFrame],
        overlay_ids: Union[dict, pandas.DataFrame],
        tileset_ids: Union[dict, pandas.DataFrame])\
        -> pandas.DataFrame:
    """
    Match IDs between game data, overlays and tileset

    >>> merge_datasets({
    ...     'type': ['ARMOR', 'TOOL', 'BOOK'],
    ...     'id': ['a', 'b', 'c'],
    ...     'description': ['desc a', 'desc b', 'desc c'],
    ...     'color': ['red', 'green', 'blue']
    ... }, {
    ...     'overlay_id': [
    ...         'overlay_worn_a', 'overlay_worn_b', 'overlay_worn_c',
    ...         'overlay_wielded_a', 'overlay_wielded_b', 'overlay_wielded_c',
    ...     ],
    ... }, {
    ...     'tileset_id': ['a', 'b', 'overlay_worn_a', 'overlay_wielded_b'],
    ... })
        type id description  color         overlay_id         tileset_id
    0  ARMOR  a      desc a    red                NaN                  a
    1  ARMOR  a      desc a    red     overlay_worn_a     overlay_worn_a
    7    NaN  a         NaN    NaN  overlay_wielded_a                NaN
    2   TOOL  b      desc b  green                NaN                  b
    3   TOOL  b      desc b  green  overlay_wielded_b  overlay_wielded_b
    5    NaN  b         NaN    NaN     overlay_worn_b                NaN
    4   BOOK  c      desc c   blue                NaN                NaN
    6    NaN  c         NaN    NaN     overlay_worn_c                NaN
    8    NaN  c         NaN    NaN  overlay_wielded_c                NaN

    # TODO:
        type id description  color         overlay_id         tileset_id
    0  ARMOR  a      desc a    red                NaN                  a
    1  ARMOR  a         NaN    NaN     overlay_worn_a     overlay_worn_a
    2  ARMOR  a         NaN    NaN  overlay_wielded_a                NaN
    3   TOOL  b      desc b  green                NaN                  b
    4   TOOL  b         NaN    NaN     overlay_worn_b                NaN
    5   TOOL  b         NaN    NaN  overlay_wielded_b  overlay_wielded_b
    6   BOOK  c      desc c   blue                NaN                NaN
    7   BOOK  c         NaN    NaN     overlay_worn_c                NaN
    8   BOOK  c         NaN    NaN  overlay_wielded_c                NaN
    """
    all_game_ids = pandas.DataFrame(all_game_ids)
    overlay_ids = pandas.DataFrame(overlay_ids)
    tileset_ids = pandas.DataFrame(tileset_ids)

    tileset_ids['id'] = tileset_ids['tileset_id'].apply(strip_overlay_id)

    # TODO: output the original ID and type in the generate_overlay_ids.py
    overlay_ids['id'] = overlay_ids['overlay_id'].apply(strip_overlay_id)
    # game_data = pandas.merge(all_game_ids, overlay_ids, on=['type', 'id'])

    # match tileset with game data
    result = all_game_ids.merge(
        tileset_ids,
        how='outer',
        on='id',
        sort=False,
    )
    # match overlays
    result = result.merge(
        overlay_ids,
        how='outer',
        left_on=['id', 'tileset_id'],
        right_on=['id', 'overlay_id'],
        sort=False,
    )
    # rearrange columns
    overlay_id_column = result.pop('overlay_id')
    result.insert(
        result.columns.get_loc('tileset_id'),
        overlay_id_column.name,
        overlay_id_column,
    )
    return result.sort_values('id')


def write_output(data: pandas.DataFrame, output_filename: str) -> int:
    """
    Write the resulting DataFrame to a file
    """
    result = data.to_csv(output_filename)
    if result is None:
        # error
        return 1

    return 0


if __name__ == '__main__':
    arg_parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    arg_parser.add_argument('all_game_ids.csv')
    arg_parser.add_argument('all_overlay_ids.csv')
    arg_parser.add_argument('tileset_ids.csv')
    arg_parser.add_argument('tileset_coverage_output.csv')
    args_dict = vars(arg_parser.parse_args())

    sys.exit(write_output(
        merge_datasets(
            *get_data(
                args_dict.get('all_game_ids.csv'),
                args_dict.get('all_overlay_ids.csv'),
                args_dict.get('tileset_ids.csv'),
            )
        ),
        args_dict.get('tileset_coverage_output.csv'),
    ))



from ..write_text import write_text


def parse_event_statistic(json, origin):
    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of event statistic \"{}\""
                   .format(json["id"]),
                   plural=True)



#!/usr/bin/env python3

# This script is intended to help monitor macOS market share to help us decide
# when it is reasonable to drop support for a particular version.
#
# The expected input file is a CSV-formatted data file downloaded from
# https://gs.statcounter.com/os-version-market-share/macos/desktop/worldwide
#
# The output shows cumulative market share by increasing version number for
# each month in the input data.  The range of versions is limited to make the
# output more reasonable, but the cumulative tally does include smaller
# versions not listed.
#
# See also doc/c++/COMPILER_SUPPORT.md

import csv
import re
import sys

in_filename, = sys.argv[1:]

with open(in_filename) as in_file:
    reader = csv.reader(in_file)
    rows = list(reader)

headers = rows[0]
rows = rows[1:]

replacement = {
    'Date': 'Date',
    'macOS Catalina': (10, 15),
    'macOS Mojave': (10, 14),
    'macOS High Sierra': (10, 13),
    'macOS Sierra': (10, 12),
    'OS X El Capitan': (10, 11),
    'OS X Mavericks': (10, 9),
    'mac OS X Snow Leopard': (10, 6),
    'mac OS X Lion': (10, 7),
    'OS X Mountain Lion': (10, 8),
    'mac OS X Leopard': (10, 6),
    'mac OS X Tiger': (10, 4),
    'mac OS X Jaguar': (10, 2),
    'mac OS X Panther': (10, 3),
    'Other': (0,)
}

version_re = re.compile('([0-9]+)\\.([0-9]+)')


def get_version(name):
    if name in replacement:
        return replacement[name]
    match = version_re.search(name)
    assert match, f'name = {name!r}'
    return tuple(int(x) for x in match.group(1, 2))


headers = [get_version(h) for h in headers]

dicts = [
    {field_name: field for field_name, field in zip(headers, row)}
    for row in rows
]

for dic in dicts:
    date = dic.pop('Date')
    stats = sorted(dic.items())
    tally = []
    total = 0
    for version, fraction in stats:
        total += float(fraction)
        if version >= (10, 11) and version <= (10, 15):
            tally.append((version, total))
    tally_str = ''
    for version, total in tally:
        tally_str += f'{version[0]}.{version[1]}: {total:4.1f}  '
    print(f'{date} :: {tally_str}')



#!/usr/bin/env python3
"""Lists duplicates in JSON by `type` and `id` fields
"""

from collections import defaultdict

from util import import_data


data = import_data()[0]
all_ids = defaultdict(set)
for obj in data:
    obj_id = obj.get('id')
    obj_type = obj.get('type')
    if obj_id and not isinstance(obj_id, list):
        if obj_id not in all_ids[obj_type]:
            all_ids[obj_type].add(obj_id)
        else:
            print(obj_type, obj_id)



#!/usr/bin/env python3
"""
Attempts to generate a reasonable set of overlay IDs
that should be defined in a tileset.
Meant to be used along with tools/gfx_tools/list_tileset_ids.py

Google Docs Spreadsheet formulas:

Prune IDs from tileset for verification:
=REGEXREPLACE(A2,"_season_(spring|summer|autumn|winter)|_male|_female","")

Search for the pruned ID in overlays
=arrayformula(iferror(VLOOKUP($C2:$C,'overlays'!$A$1:$A,1,FALSE)))

Search in game IDs
=arrayformula(iferror(VLOOKUP($C2:$C,'all game IDs'!$B$2:$B,1,FALSE)))
"""

from itertools import product

from util import import_data


CPP_IDS = (
    'cursor', 'highlight', 'highlight_item', 'footstep', 'graffiti',
    'zombie_revival_indicator',
    'weather_rain_drop', 'weather_acid_drop', 'weather_snowflake',
    'animation_bullet_normal', 'animation_bullet_shrapnel',
    'animation_bullet_flame',
    'explosion', 'explosion_weak', 'explosion_medium',
    'animation_hit', 'player_male', 'player_female', 'npc_male', 'npc_female',
    'animation_line', 'line_target', 'line_trail',
    'infrared_creature',
    'run_nw', 'run_n', 'run_ne', 'run_w', 'run_e', 'run_sw', 'run_s', 'run_se',
    'bash_complete', 'bash_effective', 'bash_ineffective',
    'shadow',
)
ATTITUDES = ('hostile', 'neutral', 'friendly', 'other')

TILESET_OVERLAY_TYPES = {
    'effect_type': {
        'prefix': 'overlay_effect_'
    },
    'mutation': {
        'prefix': 'overlay_mutation_'
    },
    'bionic': {
        'prefix': 'overlay_mutation_'
    },
    'ARMOR': {
        'prefix': 'overlay_worn_'
    },
    'TOOL_ARMOR': {
        'prefix': 'overlay_worn_'
    },
    'GENERIC': {
        'prefix': 'overlay_wielded_'
    },
    'GUN': {
        'prefix': 'overlay_wielded_'
    },
    'AMMO': {
        'prefix': 'overlay_wielded_'
    },
    'TOOL': {
        'prefix': 'overlay_wielded_'
    },
    'ENGINE': {
        'prefix': 'overlay_wielded_'
    },
    'WHEEL': {
        'prefix': 'overlay_wielded_'
    },
    'COMESTIBLE': {
        'prefix': 'overlay_wielded_'
    },
    'MONSTER': {
        'prefix': ''
    },
    'movement_mode': {
        'prefix': 'overlay_'
    },
    'vehicle_part': {
        'prefix': 'vp_'
    },
}


if __name__ == '__main__':
    data = import_data()[0]

    for datum in data:
        datum_type = datum.get('type')
        overlay_data = TILESET_OVERLAY_TYPES.get(datum_type, None)
        if not overlay_data:
            continue

        game_id = datum.get('id')
        flags = datum.get('flags', tuple())

        if not game_id:
            continue
        if datum.get('abstract'):
            continue
        if 'PSEUDO' in flags or 'NO_DROP' in flags:
            continue
        if datum.get('copy-from') in ('fake_item', 'software'):
            continue

        variable_prefix = ('',)
        variable_suffix = ('',)
        if 'BIONIC_TOGGLED' in flags:
            variable_prefix = ('', 'active_')
        if datum_type == 'MONSTER':
            variable_prefix = ('corpse_', 'overlay_wielded_corpse_')
        if datum_type == 'vehicle_part':
            variable_suffix = datum.get('symbols', {}).keys()
            if variable_suffix:
                variable_suffix = ['', ] + [f'_{s}' for s in variable_suffix]
            else:
                variable_suffix = ('',)

        for p in product(variable_prefix, (game_id,), variable_suffix):
            output = [overlay_data['prefix']]
            output.extend(p)
            # print(output)
            print(''.join(output))

    for hardcoded_id in CPP_IDS:
        print(hardcoded_id)
    for attitude in ATTITUDES:
        print(f'overlay_{attitude}_sees_player')



from ..write_text import write_text


def parse_monster_attack(json, origin):
    for key in [
        "hit_dmg_u",
        "hit_dmg_npc",
        "miss_msg_u",
        "miss_msg_npc",
        "no_dmg_msg_u",
        "no_dmg_msg_npc",
    ]:
        if key in json:
            write_text(json[key], origin, comment="Monster attack message")



#!/usr/bin/env python3

# Validate that all JSON files are syntactically correct;
# does NOT check indentation styles.

import glob
import json
import sys


def main():
    json_files = glob.glob('**/**.json', recursive=True)
    errors = 0
    total = 0
    for file_path in json_files:
        if not file_path.startswith("android/app/src/main/assets"):
            total += 1
            try:
                with open(file_path, encoding="utf-8") as fp:
                    _ = json.load(fp)
            except OSError as e:
                print("Error opening {}: {}".format(file_path, e))
                errors += 1
            except json.JSONDecodeError as e:
                print("Error parsing {}: {}".format(file_path, e))
                errors += 1
    if errors == 0:
        print("All {} JSON files healthy.".format(total))
        return 0
    else:
        print("Found {} erroneous files among {} JSON files in the repository."
              .format(errors, total))
        return min(errors, 255)


if __name__ == '__main__':
    sys.exit(main())



from ..helper import get_singular_name
from ..write_text import write_text


def parse_gunmod(json, origin):
    write_text(json["name"], origin, comment="Gun mod name", plural=True)
    name = get_singular_name(json["name"])

    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of gun mod \"{}\"".format(name))

    if "mode_modifier" in json:
        for mode in json["mode_modifier"]:
            write_text(mode[1], origin,
                       comment="Firing mode of gun mod")

    if "variants" in json:
        for variant in json["variants"]:
            variant_name = get_singular_name(variant["name"])
            write_text(variant["name"], origin,
                       comment="Variant name of gun mod \"{}\"".format(name),
                       plural=True)
            write_text(variant["description"], origin,
                       comment="Description of variant \"{1}\""
                       "of gun mod \"{0}\""
                       .format(name, variant_name))

    if "location" in json:
        write_text(json["location"], origin,
                   comment="Location of gun mod")

    if "mod_targets" in json:
        for target in json["mod_targets"]:
            write_text(target, origin, context="gun_type_type",
                       comment="Target of gun mod")



#!/usr/bin/env python3
"""Run this script with -h for usage info and docs.
"""

import argparse
import sys
import json
from util import import_data, value_counter, ui_counts_to_columns, WhereAction

parser = argparse.ArgumentParser(description="""Count the number of
times a specific values occurs for a specific key. The key may be a
single field name, or a dotted reference to name a key inside an object
or list of objects.

Example usages:

    # What values are present in the material key?
    %(prog)s --human -k material

    # What pocket_type values are used inside pocket_data values?
    %(prog)s --human -k pocket_data.pocket_type

    # What cost values are in bionics that are active?
    %(prog)s --key=cost type=bionic active=true
""", formatter_class=argparse.RawDescriptionHelpFormatter)

parser.add_argument(
    "--fnmatch",
    default="*.json",
    help="override with glob expression to select a smaller fileset.")
parser.add_argument(
    "-H", "--human",
    action="store_true",
    help="if set, makes output human readable. default is JSON output.")
parser.add_argument(
    "-L", "--list",
    action="store_true",
    help="print only a sorted list of values in JSON format,"
         " or newline-separated plain text list if --human is also set.")
parser.add_argument(
    "-k", "--key",
    required=True, type=str,
    help="key on JSON objects from which to count values")
parser.add_argument(
    "where",
    action=WhereAction, nargs='*', type=str,
    help="where exclusions of the form 'where_key=where_val', no quotes.")


if __name__ == "__main__":
    args = parser.parse_args()
    search_key = args.key

    json_data, load_errors = import_data(json_fmatch=args.fnmatch)
    if load_errors:
        # If we start getting unexpected JSON or other things, might need to
        # revisit quitting on load_errors
        print("Error loading JSON data.")
        for e in load_errors:
            print(e)
        sys.exit(1)
    elif not json_data:
        print("No data loaded.")
        sys.exit(1)

    stats, num_matches = value_counter(json_data, search_key, args.where)

    if not stats:
        print("Nothing found.")
        sys.exit(1)

    if args.human and args.list:
        print("\n".join(sorted(stats.keys())))
    elif args.human:
        title = "Count of values from field '%s'" % search_key
        print("\n\n%s" % title)
        print("(Data from %s out of %s blobs)" % (num_matches, len(json_data)))
        print("-" * len(title))
        ui_counts_to_columns(stats)
    elif args.list:
        print(json.dumps(sorted(stats.keys())))
    else:
        print(json.dumps(stats))



#!/usr/bin/env python3

# decompose.py
# Split a gfx tile_config.json into 1000s of little directories,
# each with their own config file and tile image.

"""Split a tileset's tile_config.json into one directory per tile
containing the tile data and png.
"""

import argparse
import json
import math
import os
import subprocess
import sys

try:
    vips_path = os.getenv("LIBVIPS_PATH")
    if vips_path is not None and vips_path != "":
        os.environ["PATH"] += ";" + os.path.join(vips_path, "bin")
    import pyvips
    Vips = pyvips
except ImportError:
    import gi
    gi.require_version('Vips', '8.0')
    from gi.repository import Vips


def write_to_json(pathname, data, prettify=False):
    with open(pathname, "w", encoding="utf-8") as fp:
        json.dump(data, fp, ensure_ascii=False)

    json_formatter = "./tools/format/json_formatter.cgi"
    if prettify and os.path.isfile(json_formatter):
        cmd = [json_formatter, pathname]
        subprocess.call(cmd)


def find_or_make_dir(pathname):
    try:
        os.stat(pathname)
    except OSError:
        os.mkdir(pathname)


class TileSheetData(object):
    def __init__(self, tilesheet_data, refs):
        self.ts_filename = tilesheet_data.get("file", "")
        #print("working on {}".format(self.ts_filename))
        self.tile_entries = {}
        self.sprite_height = tilesheet_data.get(
            "sprite_height", refs.default_height)
        self.sprite_width = tilesheet_data.get(
            "sprite_width", refs.default_width)
        self.sprite_offset_x = tilesheet_data.get("sprite_offset_x", 0)
        self.sprite_offset_y = tilesheet_data.get("sprite_offset_y", 0)
        self.sprite_offset_x_retracted = tilesheet_data.get(
            "sprite_offset_x_retracted",
            self.sprite_offset_x
        )
        self.sprite_offset_y_retracted = tilesheet_data.get(
            "sprite_offset_y_retracted",
            self.sprite_offset_y
        )
        self.pixelscale = tilesheet_data.get("pixelscale", 1.0)

        self.write_dim = self.sprite_width != refs.default_width
        self.write_dim |= self.sprite_height != refs.default_height
        self.write_dim |= self.sprite_offset_x or self.sprite_offset_y
        self.write_dim |= \
            self.sprite_offset_x_retracted or self.sprite_offset_y_retracted
        self.ts_pathname = refs.tileset_pathname + "/" + self.ts_filename
        self.ts_image = Vips.Image.pngload(self.ts_pathname)
        self.ts_width = self.ts_image.width
        self.ts_tiles_per_row = int(
            math.floor(self.ts_width / self.sprite_width))
        self.ts_height = self.ts_image.height
        self.ts_rows = int(math.floor(self.ts_height / self.sprite_height))
        self.pngnum_min = refs.last_pngnum
        self.pngnum_max = \
            refs.last_pngnum + self.ts_tiles_per_row * self.ts_rows - 1
        #print("\t{}x{}; {} rows; spans {} to {}".format(
        #    self.ts_width, self.ts_height, self.ts_rows,
        #    self.pngnum_min, self.pngnum_max))
        self.expansions = []
        self.fallback = tilesheet_data.get("ascii")
        refs.last_pngnum = self.pngnum_max + 1
        refs.tspathname_to_tsfilename.setdefault(
            self.ts_pathname, self.ts_filename)

    def check_for_expansion(self, tile_entry):
        if tile_entry.get("fg", -10) == 0:
            self.expansions.append(tile_entry)
            return True
        return False

    def check_id_valid(self, tile_id):
        if not tile_id:
            return True
        # wielded terrain isn't valid
        if tile_id.startswith("overlay_wielded_t_"):
            return False
        if tile_id.startswith("overlay_wielded_mon_"):
            return False
        if tile_id.startswith("overlay_wielded_fd_"):
            return False
        if tile_id.startswith("overlay_wielded_f_"):
            return False
        # wielding or wearing overlays makes no sense
        if tile_id.startswith("overlay_wielded_overlay"):
            return False
        if tile_id.startswith("overlay_worn_overlay"):
            return False
        return True

    def parse_id(self, tile_entry):
        all_tile_ids = []
        read_tile_ids = tile_entry.get("id")
        valid = True
        if isinstance(read_tile_ids, list):
            for tile_id in read_tile_ids:
                valid &= self.check_id_valid(tile_id)
                if tile_id and valid and tile_id not in all_tile_ids:
                    all_tile_ids.append(tile_id)
        else:
            valid &= self.check_id_valid(read_tile_ids)
            if read_tile_ids and valid and read_tile_ids not in all_tile_ids:
                all_tile_ids.append(read_tile_ids)
        if not valid:
            return None, None
        if not all_tile_ids:
            return "background", ["background"]
        #print("tile {}".format(all_tile_ids[0]))
        return all_tile_ids[0], all_tile_ids

    def parse_index(self, read_pngnums, all_pngnums, refs):
        local_pngnums = []
        if isinstance(read_pngnums, list):
            for pngnum in read_pngnums:
                if isinstance(pngnum, dict):
                    sprite_ids = pngnum.get("sprite", -10)
                    if isinstance(sprite_ids, list):
                        for sprite_id in sprite_ids:
                            if (sprite_id < 0 or
                                    sprite_id in refs.delete_pngnums):
                                continue
                            if sprite_id not in all_pngnums:
                                all_pngnums.append(sprite_id)
                            if sprite_id not in local_pngnums:
                                local_pngnums.append(sprite_id)
                    else:
                        if sprite_ids < 0 or sprite_ids in refs.delete_pngnums:
                            continue
                        if sprite_ids not in all_pngnums:
                            all_pngnums.append(sprite_ids)
                        if sprite_ids not in local_pngnums:
                            local_pngnums.append(sprite_ids)
                else:
                    if pngnum < 0 or pngnum in refs.delete_pngnums:
                        continue
                    if pngnum not in all_pngnums:
                        all_pngnums.append(pngnum)
                    if pngnum not in local_pngnums:
                        local_pngnums.append(pngnum)
        elif read_pngnums >= 0 and read_pngnums not in refs.delete_pngnums:
            if read_pngnums not in all_pngnums:
                all_pngnums.append(read_pngnums)
            if read_pngnums not in local_pngnums:
                local_pngnums.append(read_pngnums)
        return all_pngnums

    def parse_png(self, tile_entry, refs):
        all_pngnums = []
        fg_id = tile_entry.get("fg", -10)
        all_pngnums = self.parse_index(fg_id, all_pngnums, refs)

        bg_id = tile_entry.get("bg", -10)
        all_pngnums = self.parse_index(bg_id, all_pngnums, refs)

        add_tile_entrys = tile_entry.get("additional_tiles", [])
        for add_tile_entry in add_tile_entrys:
            add_pngnums = self.parse_png(add_tile_entry, refs)
            for add_pngnum in add_pngnums:
                if add_pngnum not in all_pngnums:
                    all_pngnums.append(add_pngnum)
        # print("\tpngs: {}".format(all_pngnums))
        return all_pngnums

    def parse_tile_entry(self, tile_entry, refs):
        if self.check_for_expansion(tile_entry):
            return None
        tile_id, all_tile_ids = self.parse_id(tile_entry)
        if not tile_id:
            # print("no id for {}".format(tile_entry))
            return None
        tile_id = tile_id.replace("/", "_")
        all_pngnums = self.parse_png(tile_entry, refs)
        offset = 0
        for i in range(0, len(all_pngnums)):
            pngnum = all_pngnums[i]
            if pngnum in refs.pngnum_to_pngname:
                continue
            pngname = "{}_{}_{}".format(pngnum, tile_id, i + offset)
            while pngname in refs.pngname_to_pngnum:
                offset += 1
                pngname = "{}_{}_{}".format(pngnum, tile_id, i + offset)
            try:
                refs.pngnum_to_pngname.setdefault(pngnum, pngname)
                refs.pngname_to_pngnum.setdefault(pngname, pngnum)
                refs.add_pngnum_to_tsfilepath(pngnum)
            except TypeError:
                print("failed to parse {}".format(
                    json.dumps(tile_entry, indent=2)))
                raise
        return tile_id

    def summarize(self, tile_info, refs):
        #debug statement to verify pngnum_min and pngnum_max
        #print("{} from {} to {}".format(
        #    self.ts_filename, self.pngnum_min, self.pngnum_max))
        if self.fallback:
            refs.ts_data[self.ts_filename] = self
            ts_tile_info = {
                "fallback": True
            }
            tile_info.append({self.ts_filename: ts_tile_info})
            return
        if self.pngnum_max > 0:
            refs.ts_data[self.ts_filename] = self
            ts_tile_info = {
                "//": "indices {} to {}".format(
                    self.pngnum_min, self.pngnum_max)
            }
            if self.write_dim:
                ts_tile_info["sprite_offset_x"] = self.sprite_offset_x
                ts_tile_info["sprite_offset_y"] = self.sprite_offset_y
                ts_tile_info["sprite_width"] = self.sprite_width
                ts_tile_info["sprite_height"] = self.sprite_height
                if self.sprite_offset_x_retracted != self.sprite_offset_x \
                        or self.sprite_offset_y_retracted \
                        != self.sprite_offset_y:
                    ts_tile_info["sprite_offset_x_retracted"] = \
                        self.sprite_offset_x_retracted
                    ts_tile_info["sprite_offset_y_retracted"] = \
                        self.sprite_offset_y_retracted
            if self.pixelscale != 1.0:
                ts_tile_info["pixelscale"] = self.pixelscale
            #print("{}: {}".format(
            #    self.ts_filename, json.dumps(ts_tile_info, indent=2)))
            tile_info.append({self.ts_filename: ts_tile_info})


class ExtractionData(object):
    def __init__(self, ts_filename, refs):
        self.ts_data = refs.ts_data.get(ts_filename)
        self.valid = False
        if not self.ts_data.sprite_width or not self.ts_data.sprite_height:
            return

        self.valid = True

        ts_base = ts_filename.split(".png")[0]
        geometry_dim = "{}x{}".format(
            self.ts_data.sprite_width, self.ts_data.sprite_height)
        pngs_dir = "/pngs_" + ts_base + "_{}".format(geometry_dim)
        self.ts_dir_pathname = refs.tileset_pathname + pngs_dir
        find_or_make_dir(self.ts_dir_pathname)
        self.tilenum_in_dir = 256
        self.dir_count = 0
        self.subdir_pathname = ""

    def write_expansions(self):
        for expand_entry in self.ts_data.expansions:
            expansion_id = expand_entry.get("id", "expansion")
            if not isinstance(expansion_id, str):
                continue
            expand_entry_pathname = "{}/{}.json".format(
                self.ts_dir_pathname, expansion_id)
            write_to_json(expand_entry_pathname, expand_entry)

    def increment_dir(self):
        if self.tilenum_in_dir > 255:
            self.subdir_pathname = "{}/images{}".format(
                self.ts_dir_pathname, self.dir_count)
            find_or_make_dir(self.subdir_pathname)
            self.tilenum_in_dir = 0
            self.dir_count += 1
        else:
            self.tilenum_in_dir += 1
        return self.subdir_pathname

    def extract_image(self, png_index, refs):
        if not png_index or refs.extracted_pngnums.get(png_index):
            return
        if png_index not in refs.pngnum_to_pngname:
            return
        pngname = refs.pngnum_to_pngname[png_index]
        ts_pathname = refs.pngnum_to_tspathname[png_index]
        ts_filename = refs.tspathname_to_tsfilename[ts_pathname]
        self.increment_dir()
        tile_data = refs.ts_data[ts_filename]
        file_index = png_index - tile_data.pngnum_min
        y_index = math.floor(file_index / tile_data.ts_tiles_per_row)
        x_index = file_index - y_index * tile_data.ts_tiles_per_row
        tile_off_x = max(0, tile_data.sprite_width * x_index)
        tile_off_y = max(0, tile_data.sprite_height * y_index)
        tile_image = tile_data.ts_image.extract_area(tile_off_x, tile_off_y,
                                                     tile_data.sprite_width,
                                                     tile_data.sprite_height)
        tile_png_pathname = self.subdir_pathname + "/" + pngname + ".png"
        tile_image.pngsave(tile_png_pathname)
        refs.extracted_pngnums[png_index] = True

    def write_images(self, refs):
        for pngnum in range(
                self.ts_data.pngnum_min, self.ts_data.pngnum_max + 1):
            out_data.extract_image(pngnum, refs)


class PngRefs(object):
    def __init__(self):
        # dict of png absolute numbers to png names
        self.pngnum_to_pngname = {}
        # dict of pngnames to png numbers; used to control uniqueness
        self.pngname_to_pngnum = {}
        # dict of png absolute numbers to tilesheet paths,
        # used to know where to extract images
        self.pngnum_to_tspathname = {}
        self.tspathname_to_tsfilename = {}
        # list of pngs written out
        self.extracted_pngnums = {}
        # list of pngs to not use
        self.delete_pngnums = []
        # misc data
        self.tileset_pathname = ""
        self.default_width = 16
        self.default_height = 16
        self.last_pngnum = 0
        self.ts_data = {}

    def get_all_data(self, tileset_dirname, delete_pathname):
        self.tileset_pathname = tileset_dirname

        try:
            os.stat(self.tileset_pathname)
        except KeyError:
            print("cannot find a directory {}".format(self.tileset_pathname))
            sys.exit(-1)

        tileset_confname = refs.tileset_pathname + "/" + "tile_config.json"

        try:
            os.stat(tileset_confname)
        except KeyError:
            print("cannot find a directory {}".format(tileset_confname))
            sys.exit(-1)

        if delete_pathname:
            with open(delete_pathname, encoding="utf-8") as del_file:
                del_ranges = json.load(del_file)
                for delete_range in del_ranges:
                    if not isinstance(delete_range, list):
                        continue
                    min_png = delete_range[0]
                    max_png = min_png
                    if len(delete_range) > 1:
                        max_png = delete_range[1]
                    for i in range(min_png, max_png + 1):
                        self.delete_pngnums.append(i)

        with open(tileset_confname, encoding="utf-8") as conf_file:
            return json.load(conf_file)

    def add_pngnum_to_tsfilepath(self, pngnum):
        if not isinstance(pngnum, int):
            return
        if pngnum in self.pngnum_to_tspathname:
            return
        for ts_filename, ts_data in self.ts_data.items():
            if pngnum >= ts_data.pngnum_min and pngnum <= ts_data.pngnum_max:
                self.pngnum_to_tspathname.setdefault(
                    pngnum, ts_data.ts_pathname)
                return
        raise KeyError("index %s out of range", pngnum)

    def convert_index(self, read_pngnums, new_index):
        if isinstance(read_pngnums, list):
            for pngnum in read_pngnums:
                if isinstance(pngnum, dict):
                    sprite_ids = pngnum.get("sprite", -10)
                    if isinstance(sprite_ids, list):
                        new_sprites = []
                        for sprite_id in sprite_ids:
                            if (sprite_id >= 0 and
                                    sprite_id not in self.delete_pngnums):
                                new_sprites.append(
                                    self.pngnum_to_pngname[sprite_id])
                        pngnum["sprite"] = new_sprites
                    else:
                        if (sprite_ids >= 0 and
                                sprite_ids not in self.delete_pngnums):
                            pngnum["sprite"] = \
                                self.pngnum_to_pngname[sprite_ids]
                    new_index.append(pngnum)
                elif pngnum >= 0 and pngnum not in self.delete_pngnums:
                    new_index.append(self.pngnum_to_pngname[pngnum])
        elif read_pngnums >= 0 and read_pngnums not in self.delete_pngnums:
            new_index.append(self.pngnum_to_pngname[read_pngnums])

    def convert_pngnum_to_pngname(self, tile_entry):
        new_fg = []
        fg_id = tile_entry.get("fg", -10)
        self.convert_index(fg_id, new_fg)
        bg_id = tile_entry.get("bg", -10)
        new_bg = []
        self.convert_index(bg_id, new_bg)
        add_tile_entrys = tile_entry.get("additional_tiles", [])
        for add_tile_entry in add_tile_entrys:
            self.convert_pngnum_to_pngname(add_tile_entry)
        tile_entry["fg"] = new_fg
        tile_entry["bg"] = new_bg

        new_id = ""
        entity_id = tile_entry["id"]
        if isinstance(entity_id, list):
            # we have to truncate the name since it could get really long,
            # which means this still may not be unique in the case of a
            # bunch instances like this but it's better than nothing...
            new_id = "_".join(entity_id)[0:150]
        else:
            new_id = entity_id

        return new_id, tile_entry

    def report_missing(self):
        for pngnum in self.pngnum_to_pngname:
            if not self.extracted_pngnums.get(pngnum):
                print("missing index {}, {}".format(
                    pngnum, self.pngnum_to_pngname[pngnum]))


if __name__ == '__main__':
    args = argparse.ArgumentParser(description=__doc__)
    args.add_argument(
        "tileset_dir", action="store",
        help="local name of the tileset directory")
    args.add_argument(
        "--delete_file", dest="del_path", action="store",
        help="local name of file containing" +
        " lists of ranges of indices to delete")
    argsDict = vars(args.parse_args())

    tileset_dirname = argsDict.get("tileset_dir", "")
    delete_pathname = argsDict.get("del_path", "")

    refs = PngRefs()
    all_tiles = refs.get_all_data(tileset_dirname, delete_pathname)

    all_tilesheet_data = all_tiles.get("tiles-new", [])
    tile_info = all_tiles.get("tile_info", {})
    if tile_info:
        refs.default_width = tile_info[0].get("width")
        refs.default_height = tile_info[0].get("height")

    overlay_ordering = all_tiles.get("overlay_ordering", [])
    ts_sequence = []
    for tilesheet_data in all_tilesheet_data:
        ts_data = TileSheetData(tilesheet_data, refs)
        ts_data.summarize(tile_info, refs)
        ts_sequence.append(ts_data.ts_filename)

    for tilesheet_data in all_tilesheet_data:
        ts_filename = tilesheet_data.get("file", "")
        ts_data = refs.ts_data[ts_filename]
        if ts_data.fallback:
            continue
        tile_entries = {}
        all_tile_entry = tilesheet_data.get("tiles", [])
        for tile_entry in all_tile_entry:
            tile_id = ts_data.parse_tile_entry(tile_entry, refs)
            if tile_id:
                tile_entries.setdefault(tile_id, [])
                tile_entries[tile_id].append(tile_entry)
        ts_data.tile_entries = tile_entries

    #debug statements to verify pngnum_to_pngname and pngname_to_pngnum
    #print("pngnum_to_pngname: {}".format(
    #    json.dumps(refs.pngnum_to_pngname, sort_keys=True, indent=2)))
    #print("pngname_to_pngnum: {}".format(
    #    json.dumps(refs.pngname_to_pngnum, sort_keys=True, indent=2)))
    #print("{}".format(json.dumps(file_tile_id_to_tile_entrys, indent=2)))
    #print("{}".format(json.dumps(refs.pngnum_to_tspathname, indent=2)))

    for ts_filename in ts_sequence:
        out_data = ExtractionData(ts_filename, refs)

        if not out_data.valid:
            continue
        out_data.write_expansions()

        for tile_id, tile_entrys in out_data.ts_data.tile_entries.items():
            #print("tile id {} with {} entries".format(
            #    tile_id, len(tile_entrys)))
            for idx, tile_entry in enumerate(tile_entrys):
                subdir_pathname = out_data.increment_dir()
                tile_entry_name, tile_entry = \
                    refs.convert_pngnum_to_pngname(tile_entry)
                if not tile_entry_name:
                    continue
                tile_entry_pathname = "{}/{}_{}.json".format(
                    subdir_pathname, tile_entry_name, idx)

                #if os.path.isfile(tile_entry_pathname):
                #    print("overwriting {}".format(tile_entry_pathname))
                write_to_json(tile_entry_pathname, tile_entry)
        out_data.write_images(refs)

    if tile_info:
        tile_info_pathname = refs.tileset_pathname + "/" + "tile_info.json"
        write_to_json(tile_info_pathname, tile_info, True)

    refs.report_missing()



from .condition import parse_condition
from ..write_text import write_text
from ..write_text import write_translation_or_var


def parse_effect(effects, origin, comment=""):
    if not type(effects) is list:
        effects = [effects]
    for eff in effects:
        if type(eff) is dict:
            if not comment:
                if "id" in eff:
                    comment = "effect \"{}\"".format(eff["id"])
                elif "effect_id" in eff:
                    comment = "effect \"{}\"".format(eff["effect_id"])
                else:
                    comment = "an effect"
            if "u_buy_monster" in eff and "name" in eff:
                write_translation_or_var(eff["name"], origin,
                                         comment="Nickname for creature '{}' "
                                         "in {}".format(eff["u_buy_monster"],
                                                        comment))
            if "snippet" not in eff or eff["snippet"] is False:
                if "message" in eff:
                    write_translation_or_var(eff["message"], origin,
                                             comment="Message in {}"
                                             .format(comment))
                if "u_message" in eff:
                    write_translation_or_var(eff["u_message"], origin,
                                             comment="Player message in {}"
                                             .format(comment))
                if "npc_message" in eff:
                    write_translation_or_var(eff["npc_message"], origin,
                                             comment="NPC message in {}"
                                             .format(comment))
                if "u_make_sound" in eff:
                    write_translation_or_var(eff["u_make_sound"], origin,
                                             comment="Player makes sound in {}"
                                             .format(comment))
                if "npc_make_sound" in eff:
                    write_translation_or_var(eff["npc_make_sound"], origin,
                                             comment="NPC makes sound in {}"
                                             .format(comment))
            if "u_roll_remainder" in eff or "npc_roll_remainder" in eff:
                if "u_roll_remainder" in eff:
                    roll_type = "player"
                else:
                    roll_type = "NPC"
                if "message" in eff:
                    write_translation_or_var(eff["message"], origin,
                                             comment="{} roll message in {}"
                                             .format(roll_type, comment))
            for cast_spell_key in ["u_cast_spell", "npc_cast_spell"]:
                if cast_spell_key not in eff:
                    continue
                spell_obj = eff[cast_spell_key]
                if "message" in spell_obj:
                    write_translation_or_var(spell_obj["message"], origin,
                                             comment="Player spell casting "
                                             "message in {}".format(comment))
                if "npc_message" in spell_obj:
                    write_translation_or_var(spell_obj["npc_message"], origin,
                                             comment="NPC spell casting "
                                             "message in {}".format(comment))
            if ("set_string_var" in eff and
                    "i18n" in eff and eff["i18n"] is True):
                str_vals = eff["set_string_var"]
                if type(str_vals) is not list:
                    str_vals = [str_vals]
                for val in str_vals:
                    write_translation_or_var(val, origin,
                                             comment="Text variable value in "
                                             "{}".format(comment))
            if "set_string_var" in eff and "string_input" in eff:
                string_input = eff["string_input"]
                if "title" in string_input:
                    write_translation_or_var(
                        string_input["title"],
                        origin,
                        comment="String input window's title in {}"
                        .format(comment)
                    )
                if "description" in string_input:
                    write_translation_or_var(
                        string_input["description"],
                        origin,
                        comment="String input window's description in {}"
                        .format(comment)
                    )
                if "default_text" in string_input:
                    write_translation_or_var(
                        string_input["default_text"],
                        origin,
                        comment="String input window's default_text in {}"
                        .format(comment)
                    )
            if ("u_spawn_monster" in eff or "npc_spawn_monster" in eff or
                    "u_spawn_npc" in eff or "npc_spawn_npc" in eff):
                if "u_spawn_monster" in eff or "npc_spawn_monster" in eff:
                    spawn_type = "monster"
                else:
                    spawn_type = "NPC"
                if "spawn_message" in eff:
                    write_text(eff["spawn_message"], origin,
                               comment="Player sees {} spawns in {}"
                               .format(spawn_type, comment))
                if "spawn_message_plural" in eff:
                    write_text(eff["spawn_message_plural"], origin,
                               comment="Player sees {}s spawn in {}"
                               .format(spawn_type, comment))
            if "u_teleport" in eff or "npc_teleport" in eff:
                if "success_message" in eff:
                    write_translation_or_var(eff["success_message"], origin,
                                             comment="Success message in "
                                             "teleportation in {}"
                                             .format(comment))
                if "fail_message" in eff:
                    write_translation_or_var(eff["fail_message"], origin,
                                             comment="Fail message in "
                                             "teleportation in {}"
                                             .format(comment))
            if "variables" in eff:
                parse_effect_variables(eff["variables"], origin, comment)
            if "run_eoc_selector" in eff:
                eoc_list = eff["run_eoc_selector"]
                if type(eoc_list) is not list:
                    eoc_list = [eoc_list]
                for eoc in eoc_list:
                    if type(eoc) is dict and "id" in eoc:
                        parse_effect_on_condition(eoc, origin,
                                                  comment="nested EOC option "
                                                  "in {}".format(comment))
                for name in eff.get("names", []):
                    write_translation_or_var(name, origin,
                                             comment="EOC option name in {}"
                                             .format(comment))
                for desc in eff.get("descriptions", []):
                    write_translation_or_var(desc, origin,
                                             comment="EOC option description "
                                             "in {}".format(comment))
                if "title" in eff:
                    write_text(eff["title"], origin,
                               comment="EOC selection title in {}"
                               .format(comment))
            if "title" in eff:
                if "u_run_inv_eocs" in eff or "u_map_run_item_eocs" in eff:
                    write_translation_or_var(eff["title"], origin,
                                             comment="Player inventory menu "
                                             "title in {}".format(comment))
                if "npc_run_inv_eocs" in eff or "npc_map_run_item_eocs" in eff:
                    write_translation_or_var(eff["title"], origin,
                                             comment="NPC inventory menu "
                                             "title in {}".format(comment))
            if "place_override" in eff:
                write_translation_or_var(eff["place_override"], origin,
                                         comment="place name in {}"
                                         .format(comment))
            # Nested effects
            if "weighted_list_eocs" in eff:
                for e in eff["weighted_list_eocs"]:
                    if type(e[0]) is dict:
                        if "effect" in e[0]:
                            parse_effect(e[0]["effect"], origin,
                                         comment="nested effect in {}"
                                         .format(comment))
            if "run_eocs" in eff:
                eoc_list = eff["run_eocs"]
                if type(eoc_list) is not list:
                    eoc_list = [eoc_list]
                for eoc in eoc_list:
                    if type(eoc) is dict and "id" in eoc:
                        parse_effect_on_condition(eoc, origin,
                                                  comment="nested EOCs in {}"
                                                  .format(comment))
            for key, cond_type in [("true_eocs", "true"),
                                   ("false_eocs", "false")]:
                if key not in eff:
                    continue
                eoc_list = eff[key]
                if type(eoc_list) is not list:
                    eoc_list = [eoc_list]
                for eoc in eoc_list:
                    if type(eoc) is not dict:
                        continue
                    parse_effect_on_condition(
                        eoc, origin,
                        comment="nested EOCs on {} condition in {}"
                        .format(cond_type, comment))
            if "if" in eff:
                parse_effect(eff["then"], origin,
                             comment="nested effect in then statement in {}"
                             .format(comment))
                if "else" in eff:
                    parse_effect(eff["else"], origin,
                                 comment="nested effect in else statement in "
                                 "{}".format(comment))
            if "switch" in eff:
                for e in eff["cases"]:
                    parse_effect(e["effect"], origin,
                                 comment="nested effect in switch statement "
                                 "in {}".format(comment))
            if "foreach" in eff:
                parse_effect(eff["effect"], origin,
                             comment="nested effect in foreach statement in {}"
                             .format(comment))


def parse_effect_on_condition(json, origin, comment=""):
    id = json["id"]
    if "effect" in json:
        if comment:
            parse_effect(json["effect"], origin,
                         comment="effect in EoC \"{}\" in {}"
                         .format(id, comment))
        else:
            parse_effect(json["effect"], origin,
                         comment="effect in EoC \"{}\"".format(id))
    if "false_effect" in json:
        if comment:
            parse_effect(json["false_effect"], origin,
                         comment="false effect in EoC \"{}\" in {}"
                         .format(id, comment))
        else:
            parse_effect(json["false_effect"], origin,
                         comment="false effect in EoC \"{}\"".format(id))
    if "condition" in json:
        parse_condition(json["condition"], origin)


def parse_effect_variables(json, origin, comment):
    if type(json) is dict:
        json = [json]
    for jsobj in json:
        for key, val in jsobj.items():
            if type(val) is dict and val.get("i18n", False):
                write_translation_or_var(val, origin,
                                         comment="value of variable '{}' in {}"
                                         .format(key, comment))



# This file is intended to be run on github CI, not locally
# It only exists because I found it way too hard to present nice
# log output *while* maintaining the correct exit status in bash.
#
# Arguably this is 3x as much code as a bash equivalent, but I
# appeal to maintanability.
#
# For local development on linux, use the following command line
# (assuming both the iwyu source root and its build dir are in the PATH):
# python iwyu_tool.py \
#     $(find src/ tests/ -maxdepth 1 -name '*.cpp' \
#           | grep -v -f tools/iwyu/bad_files.txt) \
#     -p build --jobs 4 -- \
#     -Xiwyu "--mapping_file=${PWD}/tools/iwyu/cata.imp" -Xiwyu --cxx17ns  \
#     -Xiwyu --comment_style=long -Xiwyu --max_line_length=1000

import argparse
import logging
import os
import subprocess
import sys

from pathlib import Path

# hardcoded paths that could probably be passed via command line
CHANGED_FILES_INDEX = "files_changed"
GET_AFFECTED_FILES_SCRIPT = "build-scripts/get_affected_files.py"
BLACKLIST_PATH = "tools/iwyu/bad_files.txt"
MARKER_FORCE_GLOBAL_RUN = "MARKER_CHECK_ALL"


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--iwyu_tool_path", default="iwyu_tool.py")
    args = parser.parse_args()

    print("::group::Determining files to analyze")

    # files directly changed in this PR
    changed_files = get_changed_files()
    print("changed files:")
    print_long_list(changed_files)
    # files transitively impacted by the direct change above
    # (and also the directly changed files themselves)
    affected_files = get_affected_files(changed_files)
    print("affected files:")
    print_long_list(affected_files)
    # files that we feed to IWYU. This excludes files blacklisted for
    # whatever reason
    files_to_analyze = filter_analyzable_files(affected_files)
    print("files to analyze:")
    print_long_list(files_to_analyze)
    print("::endgroup::")

    if not files_to_analyze:
        print("Nothing to analyze!")
        sys.exit(0)

    # Run IWYU with the files provided. Forward its exit code.
    status = run_iwyu_on(args.iwyu_tool_path, files_to_analyze)
    sys.exit(status)


def get_changed_files() -> list[Path]:
    # The ci workflow places the list of files changed in the PR
    # into `./files_changed` file at the root of the project.
    files_index = Path(CHANGED_FILES_INDEX)
    if not files_index.exists():
        # For pushes to master (i.e. merges) this file is *not* created.
        # We need to handle this.
        # Pretend that the core iwyu definitions changed to
        # trigger global re-check
        logging.debug(
            "no changed files index present. This is "
            "likely a push to master. Will analyze the entire codebase.")
        return [Path(MARKER_FORCE_GLOBAL_RUN)]
    paths = []
    with open(files_index) as files_index:
        for line in files_index.readlines():
            line = line.strip()
            if not line:
                continue
            paths.append(Path(line))
    return paths


def get_affected_files(changed_files: list[Path]) -> list[Path]:
    # First of all, see if any of the changed_files are global enough
    # and would require re-run on the entire codebase.
    global_files = set(Path(x) for x in [
        MARKER_FORCE_GLOBAL_RUN,
        ".github/workflows/iwyu.yml",
        "build-scripts/ci-iwyu-run.py",
        "build-scripts/get_affected_files.py",
        "tools/iwyu",  # this is a directory
        "CMakeLists.txt",
        "src/CMakeLists.txt",
        "tests/CMakeLists.txt",
    ])
    for changed in changed_files:
        if changed in global_files or any(
                parent in global_files for parent in changed.parents):
            print(
                "File %s affects global IWYU configuration so "
                "we will analyze all files" % changed)
            return generate_global_file_list()

    # Now, build-scripts/get_affected_files.py generates a list of
    # transitively affected files given a list of directly changed files.
    # This is exactly what we need.
    # Except, it requires the list of changed files to be supplied
    # to it via a file on disk, not via command-line.
    # And, if you think about it, we already have the list of changed
    # files on disk. That's `./files_changed`. So just reuse that.
    # Yes, this violates the function signature. But it is a bit less code.
    out = subprocess.run(
        [GET_AFFECTED_FILES_SCRIPT,
         "--changed-files-list", CHANGED_FILES_INDEX],
        capture_output=True, encoding="utf-8")
    if out.returncode != 0:
        print("get_affected_files.py returned with error code %d"
              % out.returncode,
              file=sys.stderr)
        print("stdout:\n  %s" % out.stdout, file=sys.stderr)
        print("stderr:\n  %s" % out.stderr, file=sys.stderr)
        sys.exit(1)
    out_paths = []
    for line in out.stdout.splitlines():
        line = line.strip()
        if not line:
            continue
        out_paths.append(Path(line))
    return out_paths


# List of all non-third-party .cpp files in the codebase.
# Equivalent to linux `find src/ tests/ -maxdepth 1 -name '*.cpp'`
def generate_global_file_list() -> list[Path]:
    paths = []
    for d in [Path("src"), Path("tests")]:
        for item in os.listdir(d):
            p = d / item
            if p.is_file() and p.suffix == ".cpp":
                paths.append(p)
    paths.sort()
    return paths


# Exclude all the files we have, well, excluded
# This is equivalent to Linux ` | grep -v -f tools/iwyu/bad_files.txt `
def filter_analyzable_files(in_files: list[Path]) -> list[Path]:
    blacklist = []
    with open(BLACKLIST_PATH, "r") as bad_files:
        for line in bad_files.readlines():
            line = line.strip()
            if line.startswith("#") or not line:
                continue
            blacklist.append(line)

    analyzable_paths = []
    for p in in_files:
        if p.suffix != ".cpp":
            continue  # because IWYU does not work on .h files
        p_str = str(p.as_posix())
        fails = any((pattern in p_str) for pattern in blacklist)
        if not fails:
            analyzable_paths.append(p)
    return analyzable_paths


def run_iwyu_on(iwyu_tool_path: str, files: list[Path]) -> int:
    argslist = [iwyu_tool_path]
    argslist.extend(str(f) for f in files)
    argslist.extend(["-p", "build", "--output-format", "clang", "--jobs", "4"])
    argslist.extend(["--"])
    cdda_root = Path(__file__).parent.parent
    mapping_path = cdda_root / "tools/iwyu/cata.imp"
    argslist.extend([
        "-Xiwyu", "--mapping_file=%s" % mapping_path,
        "-Xiwyu", "--cxx17ns",
        "-Xiwyu", "--comment_style=long",
        "-Xiwyu", "--max_line_length=1000",
        "-Xiwyu", "--error=1"])

    print("::group::IWYU full output")
    print("Running: ")
    print_long_list(argslist)
    flush_both()
    # start the process, consume its stdout, leave stderr be
    proc = subprocess.Popen(argslist, stdout=subprocess.PIPE, encoding="utf-8")
    problem_lines = []
    while True:
        line = proc.stdout.readline()
        if line == '':
            break  # IWYU finished and closed the pipe
        line = line.strip()
        print(line)
        if "#includes/fwd-decls are correct" not in line:
            problem_lines.append(line)
    proc.wait()
    flush_both()
    print("Return code ", proc.returncode)
    print("::endgroup::")

    # remove the matcher to prevent double-posting the annotations
    print("::remove-matcher owner=gcc-problem-matcher::")
    print("\n")
    if problem_lines:
        print("Problems found:")
        for line in problem_lines:
            print(line)
    elif proc.returncode == 0:
        print("No issues found!")
    else:
        print("No suggestions provided, but the process still failed somehow?")

    return proc.returncode


# GHA truncates each line to 1024 characters.
# Work around that by splitting long line into several shorter ones.
def print_long_list(things: list):
    all_lines = []
    line = ""
    for thing in things:
        thing = str(thing)
        if len(line) + len(thing) > 1000:
            all_lines.append(line)
            line = ""
        line = "%s %s" % (line, thing)
    all_lines.append(line)
    for line in all_lines:
        print("  %s" % line)


def flush_both():
    sys.stdout.flush()
    sys.stderr.flush()


if __name__ == '__main__':
    main()



from .parsers.achievement import parse_achievement
from .parsers.activity_type import parse_activity_type
from .parsers.addiction_type import parse_addiction_type
from .parsers.ammunition_type import parse_ammunition_type
from .parsers.bionic import parse_bionic
from .parsers.body_part import parse_body_part
from .parsers.character_mod import parse_character_mod
from .parsers.city import parse_city
from .parsers.climbing_aid import parse_climbing_aid
from .parsers.clothing_mod import parse_clothing_mod
from .parsers.construction import parse_construction
from .parsers.construction_category import parse_construction_category
from .parsers.construction_group import parse_construction_group
from .parsers.damage_info_order import parse_damage_info_order
from .parsers.damage_type import parse_damage_type
from .parsers.dream import parse_dream
from .parsers.effect import parse_effect_on_condition
from .parsers.effect_type import parse_effect_type
from .parsers.enchant import parse_enchant
from .parsers.end_screen import parse_end_screen
from .parsers.event_statistic import parse_event_statistic
from .parsers.faction import parse_faction
from .parsers.fault import parse_fault
from .parsers.fault_fix import parse_fault_fix
from .parsers.field_type import parse_field_type
from .parsers.furniture import parse_furniture
from .parsers.gate import parse_gate
from .parsers.generic import parse_generic
from .parsers.gun import parse_gun
from .parsers.gunmod import parse_gunmod
from .parsers.harvest import parse_harvest
from .parsers.help import parse_help
from .parsers.item_action import parse_item_action
from .parsers.item_category import parse_item_category
from .parsers.json_flag import parse_json_flag
from .parsers.keybinding import parse_keybinding
from .parsers.limb_score import parse_limb_score
from .parsers.loot_zone import parse_loot_zone
from .parsers.map_extra import parse_map_extra
from .parsers.mapgen import parse_mapgen
from .parsers.mod_migration import parse_mod_migration
from .parsers.mutation import parse_mutation
from .parsers.nested_category import parse_nested_category
from .parsers.npc import parse_npc
from .parsers.npc_class import parse_npc_class
from .parsers.option_slider import parse_option_slider
from .parsers.oter_vision import parse_oter_vision
from .parsers.overmap_terrain import parse_overmap_terrain
from .parsers.palette import parse_palette
from .parsers.profession import parse_profession
from .parsers.proficiency import parse_proficiency
from .parsers.proficiency_category import parse_proficiency_category
from .parsers.recipe import parse_recipe
from .parsers.recipe_category import parse_recipe_category
from .parsers.recipe_group import parse_recipe_group
from .parsers.magazine import parse_magazine
from .parsers.material import parse_material
from .parsers.martial_art import parse_martial_art
from .parsers.mission_definition import parse_mission_definition
from .parsers.mod_info import parse_mod_info
from .parsers.monster import parse_monster
from .parsers.monster_attack import parse_monster_attack
from .parsers.morale_type import parse_morale_type
from .parsers.movement_mode import parse_movement_mode
from .parsers.mutation_category import parse_mutation_category
from .parsers.overmap_land_use_code import parse_overmap_land_use_code
from .parsers.overmap_special import parse_overmap_special
from .parsers.practice import parse_practice
from .parsers.scenario import parse_scenario
from .parsers.shop_blacklist import parse_shopkeeper_blacklist
from .parsers.skill import parse_skill
from .parsers.skill_display_type import parse_skill_display_type
from .parsers.speech import parse_speech
from .parsers.speed_description import parse_speed_description
from .parsers.species import parse_species
from .parsers.spell import parse_spell
from .parsers.start_location import parse_start_location
from .parsers.snippet import parse_snippet
from .parsers.sub_body_part import parse_sub_body_part
from .parsers.talk_topic import parse_talk_topic
from .parsers.technique import parse_technique
from .parsers.ter_furn_transform import parse_ter_furn_transform
from .parsers.terrain import parse_terrain
from .parsers.tool_quality import parse_tool_quality
from .parsers.trap import parse_trap
from .parsers.vehicle import parse_vehicle
from .parsers.vehicle_part import parse_vehicle_part
from .parsers.vehicle_part_category import parse_vehicle_part_category
from .parsers.vehicle_spawn import parse_vehicle_spawn
from .parsers.vitamin import parse_vitamin
from .parsers.weakpoint_set import parse_weakpoint_set
from .parsers.weapon_category import parse_weapon_category
from .parsers.weather_type import parse_weather_type
from .parsers.widget import parse_widget


def dummy_parser(json, origin):
    return


parsers = {
    "achievement": parse_achievement,
    "activity_type": parse_activity_type,
    "addiction_type": parse_addiction_type,
    "ammo_effect": dummy_parser,
    "ammunition_type": parse_ammunition_type,
    "anatomy": dummy_parser,
    "ascii_art": dummy_parser,
    "attack_vector": dummy_parser,
    "behavior": dummy_parser,
    "bionic": parse_bionic,
    "bionic_migration": dummy_parser,
    "body_graph": dummy_parser,
    "body_part": parse_body_part,
    "butchery_requirement": dummy_parser,
    "camp_migration": dummy_parser,
    "character_mod": parse_character_mod,
    "charge_removal_blacklist": dummy_parser,
    "city": parse_city,
    "city_building": dummy_parser,
    "climbing_aid": parse_climbing_aid,
    "clothing_mod": parse_clothing_mod,
    "colordef": dummy_parser,
    "conduct": parse_achievement,
    "connect_group": dummy_parser,
    "construction": parse_construction,
    "construction_category": parse_construction_category,
    "construction_group": parse_construction_group,
    "damage_info_order": parse_damage_info_order,
    "damage_type": parse_damage_type,
    "dream": parse_dream,
    "disease_type": dummy_parser,
    "effect_on_condition": parse_effect_on_condition,
    "effect_type": parse_effect_type,
    "emit": dummy_parser,
    "enchantment": parse_enchant,
    "end_screen": parse_end_screen,
    "event_statistic": parse_event_statistic,
    "event_transformation": dummy_parser,
    "external_option": dummy_parser,
    "faction": parse_faction,
    "fault": parse_fault,
    "fault_fix": parse_fault_fix,
    "fault_group": dummy_parser,
    "field_type": parse_field_type,
    "furniture": parse_furniture,
    "gate": parse_gate,
    "gun": parse_gun,
    "gunmod": parse_gunmod,
    "harvest": parse_harvest,
    "harvest_drop_type": dummy_parser,
    "help": parse_help,
    "hit_range": dummy_parser,
    "item": parse_generic,
    "item_action": parse_item_action,
    "item_category": parse_item_category,
    "item_blacklist": dummy_parser,
    "item_group": dummy_parser,
    "jmath_function": dummy_parser,
    "json_flag": parse_json_flag,
    "keybinding": parse_keybinding,
    "limb_score": parse_limb_score,
    "loot_zone": parse_loot_zone,
    "magazine": parse_magazine,
    "magic_type": dummy_parser,
    "map_extra": parse_map_extra,
    "mapgen": parse_mapgen,
    "martial_art": parse_martial_art,
    "material": parse_material,
    "migration": dummy_parser,
    "mission_definition": parse_mission_definition,
    "mod_info": parse_mod_info,
    "mod_migration": parse_mod_migration,
    "mod_tileset": dummy_parser,
    "monster": parse_monster,
    "monster_adjustment": dummy_parser,
    "monster_attack": parse_monster_attack,
    "monster_blacklist": dummy_parser,
    "monster_faction": dummy_parser,
    "monster_flag": dummy_parser,
    "monster_whitelist": dummy_parser,
    "monstergroup": dummy_parser,
    "mood_face": dummy_parser,
    "morale_type": parse_morale_type,
    "movement_mode": parse_movement_mode,
    "mutation": parse_mutation,
    "mutation_category": parse_mutation_category,
    "mutation_type": dummy_parser,
    "nested_category": parse_nested_category,
    "npc": parse_npc,
    "npc_class": parse_npc_class,
    "oter_id_migration": dummy_parser,
    "oter_vision": parse_oter_vision,
    "option_slider": parse_option_slider,
    "overlay_order": dummy_parser,
    "overmap_connection": dummy_parser,
    "overmap_land_use_code": parse_overmap_land_use_code,
    "overmap_location": dummy_parser,
    "overmap_special": parse_overmap_special,
    "overmap_special_migration": dummy_parser,
    "overmap_terrain": parse_overmap_terrain,
    "palette": parse_palette,
    "playlist": dummy_parser,
    "practice": parse_practice,
    "profession": parse_profession,
    "profession_group": dummy_parser,
    "profession_item_substitutions": dummy_parser,
    "proficiency": parse_proficiency,
    "proficiency_category": parse_proficiency_category,
    "recipe": parse_recipe,
    "recipe_category": parse_recipe_category,
    "recipe_group": parse_recipe_group,
    "region_overlay": dummy_parser,
    "region_settings": dummy_parser,
    "relic_procgen_data": dummy_parser,
    "requirement": dummy_parser,
    "rotatable_symbol": dummy_parser,
    "scenario": parse_scenario,
    "scenario_blacklist": dummy_parser,
    "scent_type": dummy_parser,
    "score": dummy_parser,
    "shopkeeper_blacklist": parse_shopkeeper_blacklist,
    "shopkeeper_consumption_rates": dummy_parser,
    "skill": parse_skill,
    "skill_boost": dummy_parser,
    "skill_display_type": parse_skill_display_type,
    "snippet": parse_snippet,
    "sound_effect": dummy_parser,
    "speech": parse_speech,
    "speed_description": parse_speed_description,
    "species": parse_species,
    "spell": parse_spell,
    "start_location": parse_start_location,
    "sub_body_part": parse_sub_body_part,
    "talk_topic": parse_talk_topic,
    "technique": parse_technique,
    "temperature_removal_blacklist": dummy_parser,
    "ter_furn_transform": parse_ter_furn_transform,
    "terrain": parse_terrain,
    "ter_furn_migration": dummy_parser,
    "trait_blacklist": dummy_parser,
    "trait_group": dummy_parser,
    "trait_migration": dummy_parser,
    "trap": parse_trap,
    "trap_migration": dummy_parser,
    "tool_quality": parse_tool_quality,
    "uncraft": dummy_parser,
    "var_migration": dummy_parser,
    "vehicle": parse_vehicle,
    "vehicle_group": dummy_parser,
    "vehicle_part": parse_vehicle_part,
    "vehicle_part_category": parse_vehicle_part_category,
    "vehicle_part_migration": dummy_parser,
    "vehicle_placement": dummy_parser,
    "vehicle_spawn": parse_vehicle_spawn,
    "vitamin": parse_vitamin,
    "weakpoint_set": parse_weakpoint_set,
    "weapon_category": parse_weapon_category,
    "weather_type": parse_weather_type,
    "widget": parse_widget,
}



from ..write_text import write_text


def parse_trap(json, origin):
    id = json["id"]

    if "name" in json:
        write_text(json["name"], origin, comment="Name of a trap")
    if "vehicle_data" in json and "sound" in json["vehicle_data"]:
        write_text(json["vehicle_data"]["sound"], origin,
                   comment="Trap-vehicle collision message for trap '{}'"
                   .format(id))

    for key in ["memorial_male", "memorial_female"]:
        if key in json:
            write_text(json[key], origin,
                       comment="Memorial message of trap \"{}\"".format(id))

    if "trigger_message_u" in json:
        write_text(json["trigger_message_u"], origin,
                   comment="Message when player triggers trap \"{}\""
                   .format(id))

    if "trigger_message_npc" in json:
        write_text(json["trigger_message_npc"], origin,
                   comment="Message when NPC triggers trap \"{}\""
                   .format(id))



#!/usr/bin/env python3
"""Make a distinct list of keys from JSON input.

Main purpose to take output from other scripts, which will be JSON
dictionaries or lists of JSON dictionaries and get a distinct set of
keys from all the items.

Can be used to answer questions like, "What types of material are on
the items?" when all you want to know is types, not counts.

Example usages:

    values.py -k material | %(prog)s

"""

import argparse
import os
import stat
import json
import sys
from util import ui_values_to_columns

parser = argparse.ArgumentParser(
    description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)

parser.add_argument(
    "--human",
    action="store_true",
    help="if set, makes output human readable. default is JSON output.")


if __name__ == "__main__":
    args = parser.parse_args()

    # Saw this trick here:
    # http://stackoverflow.com/questions/13442574/how-do-i-determine-if-
    #     sys-stdin-is-redirected-from-a-file-vs-piped-from-another
    mode = os.fstat(0).st_mode
    if not stat.S_ISFIFO(mode) and not stat.S_ISREG(mode):
        # input is terminal, not what we want
        print("Redirect data to stdin, please (e.g. with a pipe)."
              " Run with -h for usage.")
        sys.exit(1)

    try:
        json_data = json.loads(sys.stdin.read())
    except Exception:
        print("Could not read json data from stdin")
        sys.exit(1)

    # We'll either get a dict or a list...
    if type(json_data) is dict:
        # ... and just make it a list.
        json_data = [json_data]

    # All keys on JSON data should be strings.
    keys = set()
    for item in json_data:
        # We assume only dictionaries here
        keys.update(list(item.keys()))

    keys = sorted(list(keys))

    if args.human:
        ui_values_to_columns(keys)
    else:
        print(json.dumps(keys))



#!/usr/bin/env python3

import argparse
import json
import os

args = argparse.ArgumentParser()
args.add_argument("dir", action="store", help="specify json directory")
args.add_argument("id", action="store", help="id of recipe to adjust")
args.add_argument("level", action="store",
                  help="what activity level to shift it to")
args_dict = vars(args.parse_args())


def gen_new(path):
    change = False
    with open(path, "r", encoding="utf-8") as json_file:
        json_data = json.load(json_file)
        for jo in json_data:
            # We only want JsonObjects
            if type(jo) is str:
                return None

            # We need a type to discriminate
            if "type" not in jo:
                return None

            # specifically, only recipes and uncrafts
            kind = jo["type"]
            if kind != "recipe" and kind != "uncraft":
                return None

            # Also, make sure it has a 'result'
            if "result" not in jo:
                return None

            # We don't want to change obsolete recipes
            if "obsolete" in jo and jo["obsolete"]:
                return None

            if jo["result"] == args_dict["id"]:
                # Already got this one
                if jo["activity_level"] != "fake":
                    print("skipping {}".format(jo["result"]) +
                          " - value is {},".format(jo["activity_level"]) +
                          " currently on {}".format(args_dict["id"]))
                    return None
                jo["activity_level"] = args_dict["level"]
                change = True

    return json_data if change else None


for root, directories, filenames in os.walk(args_dict["dir"]):
    for filename in filenames:
        path = os.path.join(root, filename)
        if path.endswith(".json"):
            new = gen_new(path)
            if new is not None:
                with open(path, "w", encoding="utf-8") as jf:
                    json.dump(new, jf, ensure_ascii=False)
                os.system(f"./tools/format/json_formatter.cgi {path}")



#!/usr/bin/env python3
'''
Creates a map folder with one of every item, furniture, terrain and trap.
Everything is spawned starting from submap 0,0.
Run this script with -h for full usage information.

Examples:

    %(prog)s data/json save/worldname
    %(prog)s data/mods/Magiclysm .
    %(prog)s data/mods/Magiclysm data/mods/Aftershock .
    %(prog)s . /tmp -f
'''

import argparse
import pathlib
import json
import math

'''
A map file is divided into 4 sections of 12x12 tiles each
'''

JSON_INFO = {
    'items': {
        'allowed_types': ['GENERIC', 'AMMO', 'MAGAZINE',
                          'MAGAZINE', 'TOOL', 'PET_ARMOR', 'BOOK',
                          'BIONIC_ITEM', 'COMESTIBLE', 'GUN', 'GUNMOD',
                          'BATTERY'],
        'tiles_per': 1
    },
    'furniture': {
        'allowed_types': ['furniture'],
        'tiles_per': 9
    },
    'terrain': {
        'allowed_types': ['terrain'],
        'tiles_per': 9
    },
    'trap': {
        'allowed_types': ['trap'],
        'tiles_per': 9
    }
}


def parse_json(path: list) -> dict:
    data_folders = []
    for data_folder in path:
        data_folder = pathlib.Path(data_folder)
        if not data_folder.exists or not data_folder.is_dir:
            return False
        data_folders.append(data_folder)

    parsed_json = dict()

    for _type in JSON_INFO:
        parsed_json[_type] = []

    for data_folder in data_folders:
        for file in data_folder.glob('**/*.json'):
            json_file = None
            with open(file, 'r') as f:
                json_file = json.load(f)

            if type(json_file) is not list:
                continue

            for _type, info in JSON_INFO.items():
                for entry in json_file:
                    if ('type' not in entry or
                        # abstract item definitions
                        'id' not in entry or
                            entry['type'] not in info['allowed_types']):
                        continue

                    parsed_json[_type].append(entry['id'])

    return parsed_json


class QuarterMap:

    def __init__(self, row, column):
        self.index = 0
        self.version = 33
        self.terrain = ['t_grass'] * 144
        self.furniture = []
        self.traps = []
        self.items = []
        self.coordinate = [row, column, 0]

    def add_item(self, itype: str) -> None:
        self.items.append({
            'row': self.index % 12,
            'column': self.index // 12,
            'typeid': itype
        })
        self.terrain[self.index] = 't_concrete'
        self.index += 1

    def add_furniture(self, ftype: str) -> None:
        index_values = (1, 4, 7, 10)
        furn_index = index_values[self.index // 4] * 12
        furn_index += index_values[self.index % 4]

        self.furniture.append({
            'row': furn_index % 12,
            'column': furn_index // 12,
            'typeid': ftype
        })

        self.terrain[furn_index] = 't_concrete'
        self.index += 1

    def add_terrain(self, ttype: str) -> None:
        index_values = (0, 3, 6, 9)
        ter_index = index_values[self.index // 4] * 12
        ter_index += index_values[self.index % 4]

        for i in range(3):
            for j in range(3):
                self.terrain[ter_index + 12 * i + j] = ttype
        self.index += 1

    def add_trap(self, trtype: str) -> None:
        index_values = (1, 4, 7, 10)
        trap_index = index_values[self.index // 4] * 12
        trap_index += index_values[self.index % 4]

        self.traps.append({
            'row': trap_index % 12,
            'column': trap_index // 12,
            'typeid': trtype
        })

        self.terrain[trap_index] = 't_concrete'
        self.index += 1

    def as_string(self, ident=0) -> str:
        identstr = '\t' * ident
        retstr = identstr + '{\n'
        retstr += identstr + f'\t"version": {self.version},\n'

        retstr += identstr + '\t"coordinates": [\n'
        retstr += identstr + f'\t\t{self.coordinate[0]},\n'
        retstr += identstr + f'\t\t{self.coordinate[1]},\n'
        retstr += identstr + f'\t\t{self.coordinate[2]}\n'
        retstr += identstr + '\t],\n'

        retstr += identstr * ident + '\t"terrain": [\n'
        for key in range(144):
            if key < len(self.terrain):
                entry = self.terrain[key]
            else:
                entry = 't_grass'
            retstr += '\t\t'
            if type(entry) is str:
                retstr += identstr + f'"{entry}"'
            else:
                retstr += identstr + f'[ "{entry[0]}", {entry[1]} ]'
            if key != 143:
                retstr += ','
            retstr += '\n'
        retstr += identstr + '\t],\n'

        retstr += identstr * ident + '\t"items": [\n'
        for key, entry in enumerate(self.items):
            retstr += '\t\t'
            retstr += identstr + \
                f'{entry["row"]}, {entry["column"]}, ' \
                f'[{{ "typeid": "{entry["typeid"]}" }}]'
            if key != len(self.items) - 1:
                retstr += ','
            retstr += '\n'
        retstr += identstr + '\t],\n'

        retstr += identstr * ident + '\t"furniture": [\n'
        for key, entry in enumerate(self.furniture):
            retstr += identstr
            retstr += "\t\t[ " + \
                f'{entry["row"]}, {entry["column"]}, ' \
                f'"{entry["typeid"]}" ]'
            if key != len(self.furniture) - 1:
                retstr += ','
            retstr += '\n'
        retstr += identstr + '\t],\n'

        retstr += identstr * ident + '\t"traps": [\n'
        for key, entry in enumerate(self.traps):
            retstr += identstr
            retstr += "\t\t[ " + \
                f'{entry["row"]}, {entry["column"]}, ' \
                f'"{entry["typeid"]}" ]'
            if key != len(self.traps) - 1:
                retstr += ','
            retstr += '\n'
        retstr += identstr + '\t]\n'

        retstr += identstr + '}'
        return retstr

    def __str__(self) -> str:
        return self.as_string()


class FullMap:

    def __init__(self, offset_x: int, offset_y: int):
        self._map = []
        self._map.append(QuarterMap(offset_x, offset_y))
        self._map.append(QuarterMap(offset_x, offset_y + 1))
        self._map.append(QuarterMap(offset_x + 1, offset_y))
        self._map.append(QuarterMap(offset_x + 1, offset_y + 1))

    def fill_items(self, items) -> None:
        i = 0
        while True:
            map_index = i // 144
            if i == 576 or i == len(items):
                break

            self._map[map_index].add_item(items[i])
            i += 1

    def fill_furniture(self, furniture) -> None:
        i = 0
        while True:
            map_index = i // 16
            if i == 64 or i == len(furniture):
                break

            self._map[map_index].add_furniture(furniture[i])
            i += 1

    def fill_terrain(self, terrain) -> None:
        i = 0
        while True:
            map_index = i // 16
            if i == 64 or i == len(terrain):
                break

            self._map[map_index].add_terrain(terrain[i])
            i += 1

    def fill_traps(self, traps) -> None:
        i = 0
        while True:
            map_index = i // 16
            if i == 64 or i == len(traps):
                break

            self._map[map_index].add_trap(traps[i])
            i += 1

    def __str__(self) -> str:
        retstr = '[\n'
        for i in range(4):
            qret = self._map[i].as_string(1)
            if i != 3:
                qret += ',\n'
            retstr += qret
        retstr += '\n]'
        return retstr


class PathNotExist(Exception):
    def __init__(self, path: pathlib.Path):
        super().__init__(f'path does not exist: {path}')


class MapsFolderExists(Exception):
    def __init__(self, path: pathlib.Path):
        super().__init__(f'maps folder already exists: {path}')


class HandlerUnhandledType(Exception):
    def __init__(self, key):
        super().__init__(f'key not handled by Handler: {key}')


class Handler:

    def __init__(self, path: pathlib.Path, parsed_ids: dict, dry_run: bool,
                 force: bool):
        self.dry_run = dry_run
        self.force = force
        self.maps_folder = pathlib.Path(path)
        if not self.maps_folder.exists():
            raise PathNotExist(self.maps_folder)
        self.maps_folder = self.maps_folder.joinpath('maps')
        if self.maps_folder.exists() and not (self.dry_run or self.force):
            raise MapsFolderExists(self.maps_folder)

        self._maps = dict()
        square_offset = 0
        for key, value in parsed_ids.items():
            if key not in JSON_INFO:
                raise HandlerUnhandledType(key)

            # 576 = 4 * 12 * 12 = Tiles in a .map file
            total_squares = math.ceil(
                JSON_INFO[key]['tiles_per'] * len(value) / 576)
            square_length = math.ceil(math.sqrt(total_squares))
            amount = 576 // JSON_INFO[key]['tiles_per']

            counter = 0
            for i in range(total_squares):
                row = i % square_length + square_offset
                column = i // square_length
                fmap = FullMap(2 * row, 2 * column)

                if key == 'items':
                    fmap.fill_items(value[counter:counter + amount])
                elif key == 'furniture':
                    fmap.fill_furniture(value[counter:counter + amount])
                elif key == 'terrain':
                    fmap.fill_terrain(value[counter:counter + amount])
                elif key == 'trap':
                    fmap.fill_traps(value[counter:counter + amount])

                self._maps[f'{row}.{column}'] = fmap
                counter += amount  # 4 x ( 12 x 12 )

            square_offset += square_length

    def save_maps(self) -> None:
        for _map in self._maps:
            row = int(_map.split('.')[0])
            column = int(_map.split('.')[1])
            folder = pathlib.Path.joinpath(self.maps_folder,
                                           f'{row // 32}.{column // 32}.0')
            if not self.dry_run:
                folder.mkdir(parents=True, exist_ok=True)

            file_path = folder.joinpath(f'{row}.{column}.0.map')
            if not self.dry_run:
                with open(file_path, 'w') as f:
                    f.write(self._maps[_map].__str__())
                    f.write('\n')
            else:
                print(file_path)
                print(self._maps[_map], end='\n\n')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument('data', nargs='+', type=str,
                        help='json files location')

    parser.add_argument(
        'path', type=str, help='folder where the maps folder will be created'
    )

    parser.add_argument('-f', '--force', action='store_true',
                        help='overwrite the maps folder if it already exists')

    parser.add_argument('-n', '--dry_run',
                        action='store_true',
                        help='print to terminal instead of creating files')

    args = parser.parse_args()
    handler = Handler(args.path, parse_json(args.data), args.dry_run,
                      args.force)
    handler.save_maps()



#!/usr/bin/env python3
from __future__ import print_function
import sys


def print_encode_error(unicode_err, counter):
    chunk = unicode_err.object
    err_line = counter + chunk.count(b'\n', 0, unicode_err.start)
    line_start = chunk.rfind(b'\n', 0, unicode_err.start) + 1
    line_end = chunk.find(b'\n', line_start)
    print("Unicode error on line {0}:".format(err_line))
    # Use RAW write because this is bytes class
    sys.stdout.buffer.write(chunk[line_start:line_end + 1])
    x_num = unicode_err.end - unicode_err.start + 2
    x_start = unicode_err.start - line_start + 2
    print("{0:>{1}}".format("^" * x_num, x_start))


def check(f):
    count = 1
    try:
        for ln in f:
            count = count + 1
    except IOError as err:
        print(err)
        return False
    except UnicodeError as UE:
        print_encode_error(UE, count)
        return False

    return True


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print("Usage: {} [FILENAME]".format(sys.argv[0]))
        sys.exit(1)
    with open(sys.argv[1], encoding="utf-8") as pot_file:
        if not check(pot_file):
            sys.exit(1)



#!/usr/bin/env python3

import argparse
import json
import os

args = argparse.ArgumentParser()
args.add_argument("dir", action="store", help="specify json directory")
args.add_argument("-v", "--verbose", action="store_true",
                  help="extra debugging output")
args.add_argument("-c", "--combine", action="store_true",
                  help="check for guns that should be combined as variants")
args.add_argument("-i", "--identifier", action="store_true",
                  help="check for guns that do not share an identifier "
                       "with their magazines")
args.add_argument("--only-good-idents", action="store_true",
                  help="if checking identifiers, output valid identifiers")
args.add_argument("-n", "--name", action="store_true",
                  help="check for guns with non-descriptive names")
args.add_argument("--no-failure", action="store_true",
                  help="do not exit as a failure if any of the checks fail")
args.add_argument("-t", "--table", action="store_true",
                  help="output collected data as a table and exit")
args_dict = vars(args.parse_args())

all_jos = dict()
VERBOSE = args_dict["verbose"]
LOADED_TYPES = {"GUN", "MAGAZINE"}
INHERITED_KEYS = [
    "id",
    "type",
    "name",
    "ammo",
    "flags",
    "volume",
    "weight",
    "modes",
    "barrel_length",
    "ranged_damage",
    "reload",
    "longest_side",
    "linkage",
    "pocket_data",
    "dispersion",
    "recoil",
    "barrel_volume",
    "blackpowder_tolerance",
]
WEIGHT_UNITS = {"kg": 1000, "g": 1}
VOLUME_UNITS = {"L": 1000, "ml": 1}
LENGTH_UNITS = {"mm": 1}
UNIT_KEYS = [
    "weight",
    "volume",
    "barrel_volume",
    "longest_side",
    "barrel_length",
]
UNIT_UNITS = {
    "weight": WEIGHT_UNITS,
    "volume": VOLUME_UNITS,
    "barrel_volume": VOLUME_UNITS,
    "longest_side": LENGTH_UNITS,
    "barrel_length": LENGTH_UNITS,
}
GUNS_BLACKLIST = {
    # Well named, faction-specific weapon, with weird ammo
    "brogyeki",
    "khuunaofaai",
    "exodii_plasma_fan",
    "exodii_plasma_projectile",
}
AMMO_BLACKLIST = {
    "arrow",
    "bolt",
    "payload",
    "chemical_spray",
    "fishspear",
    "nail",
    "paintball",
    "signal_flare",
    "water",
}
VARIANT_CHECK_BLACKLIST = {
    "pamd68rubik",

}
VARIANT_CHECK_PAIR_BLACKLIST = {
    # FIXME: fix and remove these
    ("m17", "m18"),
    ("pamd68", "pamd68mountable"),
    ("type99", "type99_sniper"),
    ("glock_20", "glock_40"),
}
IDENTIFIER_CHECK_BLACKLIST = {
    # FIXME: fix and remove these
    "bigun",
    "american_180",
    "fn_fal_semi",
    "m1a",
    "rfb_308",
    "steyr_scout",
    "mdrx",
    "STI_DS_10",
    "hpt3895",
    "m2carbine",
}
NAME_CHECK_BLACKLIST = {
    # FIXME: fix and remove these
    "fn_p90",
    "hk_mp7",
    "obrez",
    "pressin",
    "m2010",
    "weatherby_5",
    "win70",
    "mr73",
    "iwi_tavor_x95_300blk",
    "sig_mcx_rattler_sbr",
    "p226_357sig",
    "glock_31",
    "p320_357sig",
    "scar_l",
    "brogyaga",
    "raging_bull",
    "raging_judge",
    "saiga_410",
    "shotgun_410",
    "mgl",
    "pseudo_m203",
    "atgm_launcher",
    "xedra_gun",
    "90two40",
    "glock_22",
    "px4_40",
    "sig_40",
    "hi_power_40",
    "walther_ppq_40",
    "hptjcp",
    "AT4",
    "af2011a1_38super",
    "m1911a1_38super",
    "plasma_gun",
    "bbgun",
}
# Stripped from gun/mag names before checking for an identifier
BAD_IDENTIFIERS = [
    "10mm",
    ".22",
    ".338",
    ".38",
    ".40",
    ".44",
    ".45",
    ".500",
    "5x50mm",
    "5.7mm",
    "7.7mm",
    "8x40mm",
    "9x19mm",
    "-round",
    "magazine",
    "pistol",
    "stripper",
    "speedloader",
]
# Common tokens that will be rejected
BAD_COMMON_TOKENS = {
    "rifle",
    "carbine",
    "pistol",
    "ing"
}
# Common tokens that are permitted to be below length reqs
SHORT_COMMON_TOKENS = {
    "AI",  # Abbreviation of gun manufacturer
    "AK",  # Common name for a gun family
    "FN",  # Common name for a manufacturer
    "G3",  # Common name for a certain gun
    "M9",  # Common name for a certain gun
}
TYPE_DESCRIPTORS = [
    "automagnum",
    "blunderbuss",
    "cannon",
    "carbine",
    "coilgun",
    # Not great, but weird can get a pass
    "combination gun",
    "derringer",
    "flamethrower",
    "flintlock",
    # Special faction-specific invented weapons get a pass
    "FSP",
    "gatling gun",
    "hand cannon",
    "handgun",
    "HMG",
    # Special faction-specific invented weapons get a pass
    "HWP",
    "launcher",
    "lever gun",
    "LMG",
    "machine gun",
    "minigun",
    # This is as close as you can get without a giant name
    "operational briefcase",
    "pistol",
    "railgun",
    "revolver",
    "rifle",
    "shotgun",
    "six-shooter",
    "SMG",
    "submachine gun",
    "trenchgun",
]
DUPE_CHECK_BLACKLIST = [
    # No magazines, and very similar tube-fed rifles with slightly diff barrels
    {"rio_bravo", "henry_golden_boy"},
]

"""
JSON parsing functions
"""


# Helper to parse quantities
def parse_unit(val, stops):
    # already parsed
    if type(val) is not str:
        return val
    total = 0
    unit = 0
    sign = 1
    unit_start = -1
    eaten = 0
    while len(val[eaten:]) != 0:
        if val[eaten].isdigit():
            unit *= 10
            unit += ord(val[eaten]) - ord('0')
        elif val[eaten] == '-':
            sign = -1
        # Non-space, non-decimal characters must be part of a unit
        elif not val[eaten].isspace():
            unit_start = eaten if unit_start == -1 else unit_start
            # We found a valid unit
            if val[unit_start:eaten + 1] in stops:
                unit *= stops[val[unit_start:eaten + 1]]
                total += unit * sign
                unit = 0
                sign = 1
                unit_start = -1
        eaten += 1
    return total


# Parse the gun data represented in JSON to what we care about
def simplify_gun_data(jo):
    # We only care about what modes it has, and how many bullets
    if "modes" in jo:
        modes = []
        for mode in jo["modes"]:
            modes.append(mode[2])
        jo["modes"] = sorted(modes)
    else:
        jo["modes"] = [1]
    # We only care how much damage it deals
    if "ranged_damage" in jo:
        if type(jo["ranged_damage"]) is not dict:
            raise ValueError("Multiple damage types not supported")
        jo["ranged_damage"] = jo["ranged_damage"]["amount"]
    else:
        jo["ranged_damage"] = 0
    # And we only care what magazines it takes (or how much ammo, if no mags)
    # Oh god why
    if "pocket_data" not in jo:
        if VERBOSE:
            print("\tRemoving %s, conversion kit gun" % jo["id"])
        return False
    for data in jo["pocket_data"]:
        # It does not take magazines
        if data["pocket_type"] == "MAGAZINE":
            # First value in dict
            quantity = next(iter(data["ammo_restriction"].values()))
            # Check that ammo restriction is uniform
            if (len(data["ammo_restriction"]) != 1 and
               len(set(data["ammo_restriction"].values())) != 1):
                raise ValueError("Only one quantity allowed for ammo")
            # TODO: figure out a better way to specify this
            jo["magazines"] = [quantity]
            # TODO: combine with magazines?
            if "allowed_speedloaders" in data:
                jo["speedloaders"] = data["allowed_speedloaders"]
        # It takes magazines (so much simpler!)
        elif data["pocket_type"] == "MAGAZINE_WELL":
            jo["magazines"] = sorted(data["item_restriction"])
        else:
            raise ValueError("Guns with other pockets not supported")
    # We don't need all of this data
    del jo["pocket_data"]
    return True


# Remove irrelevant keys and clean data to ease later checks
def simplify_object(jo):
    # Only guns are subject to intense scrutiny
    # everything else can be considered already simple
    if jo["type"] != "GUN":
        return True
    if "id" not in jo or jo["id"] in GUNS_BLACKLIST:
        if VERBOSE:
            name = str(jo)
            if "id" in jo:
                name = jo["id"]
            if "abstract" in jo:
                name = jo["abstract"]
            print("\tRemoving %s, blacklisted or abstract" % name)
        return False

    req_keys = {"weight", "volume", "ammo", "id"}
    all_keys = req_keys | set(INHERITED_KEYS)

    # Drop all the other keys
    removed = list(filter(lambda key: key not in all_keys, jo.keys()))
    # Need to iterate over removed because we can't delete from dict in for
    for key in removed:
        del jo[key]

    # guns without the required keys can't be checked
    for key in filter(lambda key: key not in jo, req_keys):
        if VERBOSE:
            print("\tRemoving %s, lacking %s" % (jo["id"], key))
        return False
    # We don't care about fake guns
    if "flags" in jo and "PSEUDO" in jo["flags"]:
        if VERBOSE:
            print("\tRemoving %s, pseudo gun" % jo["id"])
        return False

    return simplify_gun_data(jo)


# Given a JSON object, extract the player-visible name
def name_of(jo):
    if type(jo["name"]) is dict:
        if "str" in jo["name"]:
            return jo["name"]["str"]
        else:
            return jo["name"]["str_sp"]
    return jo["name"]


"""
JSON loading functions
"""


def extract_jos(path):
    with open(path, "r", encoding="utf-8") as json_file:
        try:
            json_data = json.load(json_file)
        except json.JSONDecodeError:
            print(f"Json Decode Error at {path}")
            print("Ensure that the file is a JSON file consisting of"
                  " an array of objects!")
            return None

        for jo in json_data:
            if type(jo) is not dict:
                print(f"Incorrectly formatted JSON data at {path}")
                print("Ensure that the file is a JSON file consisting"
                      " of an array of objects!")
                break

            if "type" not in jo or jo["type"] not in LOADED_TYPES:
                continue

            ident = jo["id"] if "id" in jo else jo["abstract"]
            all_jos[ident] = jo


def copy_from_extend(jo, data):
    if "extend" not in jo:
        return
    subobject = jo["extend"]
    for key in INHERITED_KEYS:
        if key not in subobject:
            continue
        for val in subobject[key]:
            if key not in data:
                data[key] = [val]
            else:
                data[key].append(val)


def copy_from_delete(jo, data):
    if "delete" not in jo:
        return
    subobject = jo["delete"]
    for key in INHERITED_KEYS:
        if key not in subobject:
            continue
        for val in subobject[key]:
            if key not in data:
                raise ValueError("Invalid delete on %s" % key)
            else:
                data[key].remove(val)


def copy_from_relative(jo, data):
    if "relative" not in jo:
        return
    subobject = jo["relative"]
    for key in INHERITED_KEYS:
        if key not in subobject:
            continue
        if key not in data:
            raise ValueError("Invalid relative on %s" % key)
        if key in UNIT_KEYS:
            data[key] += parse_unit(subobject[key], UNIT_UNITS[key])
        elif (key == "ranged_damage" and
              type(data[key]) is dict and
              type(subobject[key]) is dict):
            data[key]["amount"] += subobject[key]["amount"]
        elif key in {"dispersion"}:
            data[key] += subobject[key]
        else:
            raise ValueError("Relative for %s not supported" % key)


def copy_from_proportional(jo, data):
    if "proportional" not in jo:
        return
    subobject = jo["proportional"]
    for key in INHERITED_KEYS:
        if key not in subobject:
            continue
        if key not in data:
            raise ValueError("Invalid proportional on %s" % key)
        if (key == "ranged_damage" and
           type(data[key]) is dict and
           type(subobject[key]) is dict):
            data[key]["amount"] *= subobject[key]["amount"]
        elif key in UNIT_KEYS or key in {"reload", "dispersion"}:
            data[key] *= subobject[key]
        else:
            raise ValueError("Proportional for %s not supported" % key)


def do_copy_from(jo):
    data = {}
    if "copy-from" in jo and jo["copy-from"] in all_jos:
        data = do_copy_from(all_jos[jo["copy-from"]])

    for key in INHERITED_KEYS:
        if key in jo:
            if key in UNIT_KEYS:
                data[key] = parse_unit(jo[key], UNIT_UNITS[key])
            elif key == "name":
                data[key] = name_of(jo)
            else:
                data[key] = jo[key]

    copy_from_extend(jo, data)
    copy_from_delete(jo, data)
    copy_from_relative(jo, data)
    copy_from_proportional(jo, data)

    return data


def load_all_json(directory):
    for root, directories, filenames in os.walk(directory):
        for filename in filenames:
            if not filename.endswith(".json"):
                continue
            path = os.path.join(root, filename)
            extract_jos(path)

    for gun in all_jos.keys():
        new = do_copy_from(all_jos[gun])
        all_jos[gun] = new


"""
CHECK 1: Should a gun be a variant?
"""


# Are these guns similar enough they should be variants?
def guns_are_similar(jo, gun):
    # For volume, weight, length, etc
    # Compare if they're in the same or an adjacent bucket
    def in_range(lhs, rhs, bucket_size):
        return abs((lhs // bucket_size) - (rhs // bucket_size)) < 2

    def in_pct(lhs, rhs, pct):
        low = 1 - pct
        high = 1 + pct
        return ((lhs >= rhs * low and lhs <= rhs * high) or
                (rhs >= lhs * low and rhs <= lhs * high))

    # "symmetric difference" of sets - check if they completely intersect
    def sets_match(lhs, rhs):
        return len(set(lhs) ^ set(rhs)) == 0

    if (gun["id"] in VARIANT_CHECK_BLACKLIST or
       jo["id"] in VARIANT_CHECK_BLACKLIST):
        return False
    if not sets_match(jo["ammo"], gun["ammo"]):
        return False
    if not sets_match(jo["modes"], gun["modes"]):
        return False
    if not in_range(jo["volume"], gun["volume"], 250):
        return False
    if not in_range(jo["weight"], gun["weight"], 250):
        return False
    if not in_range(jo["ranged_damage"], gun["ranged_damage"], 2):
        return False
    # If longest side isn't specified for either, don't care about it
    if ("longest_side" in jo and "longest_side" in gun and
       not in_range(jo["longest_side"], gun["longest_side"], 50)):
        return False
    # Something to check if dispersion is similar
    # This can be pretty niche, mostly matters for handcrafted guns
    if not in_pct(jo["dispersion"] if "dispersion" in jo else 0,
                  gun["dispersion"] if "dispersion" in gun else 0, 0.05):
        return False
    # If one has speedloaders but the other doesn't
    if ("speedloaders" in jo) != ("speedloaders" in gun):
        return False
    # If both have different speedloaders
    if ("speedloaders" in jo and "speedloaders" in gun and
       not sets_match(jo["speedloaders"], gun["speedloaders"])):
        return False
    if not sets_match(jo["magazines"], gun["magazines"]):
        return False
    if not in_range(jo["reload"] if "reload" in jo else 0,
                    gun["reload"] if "reload" in gun else 0,
                    125):
        return False
    if ("barrel_length" in jo and "barrel_length" in gun and
        # this can be super sensitive to small changes in mm
        # at the low end, the bucket size is ~1-to-1 with damage
        # ranges that can be combined. Most things won't be
        # at the low end, and it seems to middle ranges of size
        # have a 10-20 mm separation per damage rather than 2-3
        # TODO: some sort of statistical analysis based on the ammo
        #       and DT of gun determines bucket size?
       not in_range(jo["barrel_length"], gun["barrel_length"], 10)):
        return False
    # TODO: get values for comparing these
    #if ("barrel_volume" in jo and "barrel_volume" in gun and
    #    not in_range(jo["barrel_volume"], gun["barrel_volume"], 125)):
    #    return False
    #if ("" in jo and "blackpower_tolerance" in gun and
    #    not in_range(jo["blackpower_tolerance"],
    #                 gun["blackpower_tolerance"], 125)):
    #    return False
    #if ("recoil" in jo and "recoil" in gun and
    #    not in_range(jo["recoil"], gun["recoil"], 125)):
    #    return False
    # TODO: more with flags?
    return True


def find_variants(all_guns):
    similar_guns = []
    for i in range(len(all_guns)):
        gun = all_guns[i]
        # We've already checked this gun against the guns before it,
        # so only check against guns after
        for compared in all_guns[i + 1:]:
            lid = gun["id"]
            rid = compared["id"]
            # We expect the same gun to be similar
            if (guns_are_similar(gun, compared) and
                not ((lid, rid) in VARIANT_CHECK_PAIR_BLACKLIST or
                     (rid, lid) in VARIANT_CHECK_PAIR_BLACKLIST)):
                similar_guns.append((gun, compared))
    return similar_guns


"""
CHECK 2: Are the default gun names linking the gun and it's mags
"""


# https://stackoverflow.com/questions/2892931/longest-common-substring-from-more-than-two-strings/2894073#2894073
# Used to find the shared token
def longest_common_substring(dat):
    substr = ''
    if len(dat) > 1 and len(dat[0]) > 0:
        for i in range(len(dat[0])):
            for j in range(len(dat[0]) - i + 1):
                if j > len(substr) and all(dat[0][i:i + j] in x for x in dat):
                    substr = dat[0][i:i + j]
    return substr


# Remove common but non-meaningful tokens from names
def clean_names(names):
    ret = []
    for name in names:
        for token in BAD_IDENTIFIERS:
            name = name.replace(token, "")
        ret.append(name)
    return ret


# Find the common token with a gun and the mags it takes
def common_token(names):
    names = clean_names(names)
    # Assume the longest common substring will be the "identifier"
    # leading/trailing whitespace isn't meaningful in identifiers
    common_token = longest_common_substring(names).strip()
    # It can't be a meaningful identifier if it's 1-2 characters long
    # Some exceptions (e.g. AK, G3)
    if len(common_token) < 3 and common_token not in SHORT_COMMON_TOKENS:
        return None
    # Some common identifiers that don't work
    if common_token in BAD_COMMON_TOKENS:
        return None
    return common_token


# Print the identifier info for a gun and it's magazines/speedloaders
def print_identifier_info(token, gun, magazines):
    mags = ""
    alt_mags = ""
    for mag in magazines:
        name = name_of(all_jos[mag]) if mag in all_jos else ""
        # Skip STANAG names
        if "STANAG" in name:
            alt_mags += mag + ", "
            continue
        mags += mag + f" ({name})" + ", "
    print(token)
    print("  " + gun["id"] + " (" + name_of(gun) + ")")
    if (len(mags) > 0):
        print("  " + mags)
    if (len(alt_mags) > 0):
        print("  " + alt_mags)


def find_identifiers(all_guns):
    # The guns with good identifiers linking the gun and it's magazines
    good_tokens = []
    # The guns with nothing in common or otherwise poor identifiers
    bad_tokens = []
    for gun in all_guns:
        if gun["id"] in IDENTIFIER_CHECK_BLACKLIST:
            continue
        # The names of the gun and all the magazines it accepts
        names = [name_of(gun)]
        # The identifiers of the magazines it accepts
        mags = []
        for mag in gun["magazines"]:
            if mag not in all_jos:
                if VERBOSE and not type(mag) is not str:
                    print("\tnot checking magazine %s for %s" %
                          (mag, gun["id"]))
                continue
            mags.append(mag)
        # Add all the magazine names in
        for mag in mags:
            name = name_of(all_jos[mag])
            # Skip STANAG, because those are uber generic, also linkages
            if "STANAG" not in name and "linkage" not in all_jos[mag]:
                names.append(name)
        # If it didn't have any mags, we can skip it
        if len(names) < 2:
            continue
        # Check if they've got a good token
        token = common_token(names)
        if token is None:
            bad_tokens.append((gun, mags))
        else:
            good_tokens.append((gun, mags, token))
    return good_tokens, bad_tokens


"""
CHECK 3: names say what they are, and aren't just gobbledygook
"""


def find_bad_names(all_guns):
    bad_names = []
    for gun in all_guns:
        if gun["id"] in NAME_CHECK_BLACKLIST:
            continue
        name = name_of(gun)
        good = False
        for type_descriptor in TYPE_DESCRIPTORS:
            if type_descriptor.upper() in name.upper():
                good = True
                break
        if not good:
            bad_names.append((gun["id"], name))
    return bad_names


"""
CHECK 4: No dupes!
"""


def remove_blacklisted_dupes(guns):
    if len(guns) == 1:
        return guns
    guns_set = set(map(lambda gun: gun["id"], guns))
    for blacklist_set in DUPE_CHECK_BLACKLIST:
        all_match = True
        for gun in blacklist_set:
            if gun not in guns_set:
                print(gun, "not in", guns_set)
                all_match = False
                break
        if all_match:
            guns = list(filter(lambda g: g["id"] not in blacklist_set, guns))
    return guns


def find_dupe_names(all_guns):
    error = False
    all_names = {}
    for gun in all_guns:
        name = name_of(gun) + str(sorted(gun["ammo"]))
        if name in all_names:
            all_names[name].append(gun)
        else:
            all_names[name] = [gun]
    for key, value in all_names.items():
        value = remove_blacklisted_dupes(value)
        if len(value) < 2:
            continue
        out = "ERROR: Guns have the same name and ammo (" + key + "):"
        error = True
        for gun in value:
            out += " (" + gun["id"] + "),"
        print(out)
    return error


"""
Driver functions
"""


# True on failure
def check_combination(all_guns):
    if not args_dict["combine"]:
        return False

    print("==== VARIANT COMBINATION ====")
    similar_guns = find_variants(all_guns)
    for pair in similar_guns:
        print("ERROR: Guns %s and %s are too similar and should be combined" %
              (pair[0]["id"], pair[1]["id"]))

    return len(similar_guns) > 0


def string_listify(strings, separator):
    count = len(strings)
    if count == 0:
        return ""
    elif count == 1:
        return strings[0]

    ret = ""
    for i in range(count - 1):
        ret += strings[i] + separator
    ret += strings[count - 1]
    return ret


def check_identifiers(all_guns):
    if not args_dict["identifier"]:
        return False

    print("====  IDENTIFIER  CHECK  ====")
    good_tokens, bad_tokens = find_identifiers(all_guns)
    if args_dict["only_good_idents"]:
        for good in good_tokens:
            print_identifier_info(good[2], good[0], good[1])
    else:
        for bad in bad_tokens:
            error_str = "This gun and it's magazines lack a shared identifier:"
            print_identifier_info(f"ERROR: {error_str}", bad[0], bad[1])
        if VERBOSE:
            good_token_list = {}
            for good in good_tokens:
                if good[2] in good_token_list:
                    good_token_list[good[2]].append(good[0]["id"])
                else:
                    good_token_list[good[2]] = [good[0]["id"]]
            # make sense: are real words (or gun names (: ), when they apply to
            # multiple guns, those guns all take the same mags, etc
            print("The following valid identifiers were found.",
                  "Please check to ensure they make sense.")
            good_tokens = [[]]
            idx = 0
            len_so_far = 0
            for token in sorted(good_token_list):
                guns_str = string_listify(good_token_list[token], " ")
                good_tokens[idx].append(f"{token} ({guns_str})")
                len_so_far += len(good_tokens[idx][-1])
                if len_so_far > 100:
                    len_so_far = 0
                    idx += 1
                    good_tokens.append([])
            for string in good_tokens:
                print(" -", string_listify(string, ", "))

    return len(bad_tokens) > 0


def check_names(all_guns):
    if not args_dict["name"]:
        return False
    print("====  DESCRIPTIVE NAMES  ====")
    bad_names = find_bad_names(all_guns)
    for bad in bad_names:
        print("ERROR: Gun %s (%s) lacks a descriptive name" % (bad[1], bad[0]))

    dupes = find_dupe_names(all_guns)

    return len(bad_names) > 0 or dupes


csv = True


def insert_start_char():
    return "" if csv else "| "


def insert_separator():
    return "," if csv else " | "


def table(all_guns):
    # Don't report these
    skipped = {
        "type",
        "barrel_volume",
        "blackpowder_tolerance",
        "pocket_data",
        "recoil"
    }
    # If we don't have these, report this
    defaults = {
        "flags": [],
        "barrel_length": "???",
        "reload": 0,
        "longest_side": "???",
        "speedloaders": [],
        "dispersion": "???"
    }
    # Do the inherited keys, except the skipped ones, plus these
    keys = INHERITED_KEYS + ["magazines", "speedloaders"]
    # Print a header
    out = insert_start_char()
    for key in keys:
        if key in skipped:
            continue
        out += key
        out += insert_separator()
    print(out)
    # Markdown needs an extra line
    if not csv:
        out = insert_start_char()
        for key in keys:
            if key in skipped:
                continue
            out += "----"
            out += insert_separator()
        print(out)
    # Print all the gun data
    for gun in all_guns:
        out = insert_start_char()
        for key in keys:
            if key in skipped:
                continue
            if key in gun and type(gun[key]) is list:
                for val in gun[key]:
                    out += "'" + str(val) + "' "
            else:
                out += str(gun[key]) if key in gun else str(defaults[key])
            out += insert_separator()
        print(out)


def main():
    # Load all the JSON for the types we care about
    load_all_json(args_dict["dir"])
    # Find all the guns and simplify them
    all_guns = []
    for jo in all_jos.values():
        if jo["type"] != "GUN":
            continue
        if ("flags" in jo and
            ("PRIMITIVE_RANGED_WEAPON" in jo["flags"] or
             "PSEUDO" in jo["flags"])):
            continue
        if not simplify_object(jo):
            continue
        # blacklisted ammo
        if len(set(jo["ammo"]).intersection(AMMO_BLACKLIST)) != 0:
            continue
        all_guns.append(jo)
    if args_dict["table"]:
        table(all_guns)
        os.sys.exit(0)

    failed = False
    if check_combination(all_guns):
        failed = True
    if check_identifiers(all_guns):
        failed = True
    if check_names(all_guns):
        failed = True

    if failed and not args_dict["no_failure"]:
        os.sys.exit(1)
    else:
        os.sys.exit(0)


if __name__ == "__main__":
    main()



from ..write_text import write_text


def parse_item_category(json, origin):
    if "name_header" in json:
        write_text(json["name_header"], origin,
                   comment="Item category name used for headers")
    if "name_noun" in json:
        write_text(json["name_noun"], origin,
                   comment="Item category name used for descriptive text",
                   plural=True)



from ..helper import get_singular_name
from ..write_text import write_text


def parse_faction(json, origin):
    name = ""
    if "name" in json:
        name = get_singular_name(json["name"])
        write_text(json["name"], origin, comment="NPC faction name")

    if "description" in json:
        write_text(json["description"], origin,
                   comment="Description of NPC faction \"{}\"".format(name))



#!/usr/bin/env python3

import json
import argparse

IGNORE_MISMATCH = ["id", "abstract"]


def get_data(argsDict, resource_name):
    resource = []
    resource_sources = argsDict.get(resource_name, [])
    if not isinstance(resource_sources, list):
        resource_sources = [resource_sources]
    for resource_filename in resource_sources:
        if resource_filename.endswith(".json"):
            try:
                with open(
                        resource_filename, encoding="utf-8") as resource_file:
                    resource += json.load(resource_file)
            except FileNotFoundError:
                exit("Failed: could not find {}".format(resource_filename))
        else:
            print(("Invalid filename {}".format(resource_filename)))
    if not resource:
        exit("Failed: {} was empty".format(resource_filename))
    return resource


args = argparse.ArgumentParser(description="Make items use copy-from.")
args.add_argument("item_source", action="store",
                  help="specify json file to rewrite using copy-from.")
args.add_argument("--output", dest="output_name", action="store",
                  help="Name of output file.  Defaults to the command line.")
argsDict = vars(args.parse_args())

items = get_data(argsDict, "item_source")

base_item = items[0]
base_id = base_item.get("id") or base_item.get("abstract")
if not base_id:
    exit("Failure: first item malformed")
for item in items:
    if item == base_item:
        continue
    if item.get("copy-from"):
        continue
    del_keys = []
    for key in item:
        if key == "type":
            continue
        if item[key] == base_item.get(key):
            del_keys.append(key)
    can_copy = True
    for key in base_item:
        if key in del_keys:
            continue
        if key in IGNORE_MISMATCH:
            continue
        if key not in item:
            can_copy = False
            break

    if not can_copy:
        continue

    for key in del_keys:
        del item[key]
    item["copy-from"] = base_id

output_name = argsDict.get("output_name", "")
if output_name and not output_name.endswith(".json"):
    output_name += ".json"

print(json.dumps(items, indent=2))
